package space.uranos.mdsp.generator

import com.google.common.base.CaseFormat
import com.jsoniter.JsonIterator
import com.squareup.kotlinpoet.*
import com.squareup.kotlinpoet.ParameterizedTypeName.Companion.parameterizedBy
import space.uranos.mdsp.JsonAny
import java.io.File

class BlocksAndMaterialGenerator(
    private val workingDir: File,
    private val outputDir: File,
    private val sourcesDir: File
) {
    companion object {
        val BLOCK_BLACKLIST = listOf(
            // Only used on the client side for y values lower than 0 and higher than 255
            "void_air"
        )
    }

    fun generate() {
        val collisionShapesJson = workingDir.resolve("blockCollisionShapes.json").readText()
        val collisionShapesData = JsonIterator.deserialize(collisionShapesJson)
        generateCollisionShapes(collisionShapesData.get("shapes").asMap())

        val dataJson = workingDir.resolve("blocks.json").readText()
        val blocks = JsonIterator.deserialize(dataJson).asList()

        generateBlockStubs(blocks, collisionShapesData.get("blocks").asMap())
        generateBlocksList(blocks)
    }

    private fun generateCollisionShapes(shapes: Map<String, JsonAny>) {
        val items = shapes.toList().sortedBy { it.first.toInt() }.map { shape ->
            "arrayOf(${
                shape.second.asList().joinToString { outer ->
                    "arrayOf(${outer.joinToString { inner -> inner.toString() + "f" }})"
                }
            })"
        }

        val typeAlias = TypeAliasSpec.builder(
            "CollisionShape",
            Array::class.asTypeName().parameterizedBy(Array::class.asTypeName().parameterizedBy(ClassName("kotlin", "Float")))
        ).build()

        val property = PropertySpec.builder(
            "BLOCK_COLLISION_SHAPES",
            Array::class.asTypeName().parameterizedBy(ClassName(BLOCK_PACKAGE, "CollisionShape"))
        )
            .initializer("arrayOf(${items.joinToString()})")
            .addKdoc(
                "Explanation what the numbers mean: " +
                    "https://github.com/PrismarineJS/minecraft-data/blob/master/doc/blockCollisionShapes.md"
            )
            .build()

        FileSpec.builder(BLOCK_PACKAGE, "CollisionShapes")
            .addTypeAlias(typeAlias)
            .addProperty(property)
            .build()
            .writeTo(outputDir)
    }

    private val materialLines = listOf("%L", "%S", "%L", "%Lf", "%L", "%L", "%L", "BLOCK_COLLISION_SHAPES[%L]")

    private fun generateBlockStubs(blocks: List<JsonAny>, collisionShapes: Map<String, JsonAny>) {
        for (block in blocks) {
            val lowerUnderscoreName = block.get("name").toString()
            if (BLOCK_BLACKLIST.contains(lowerUnderscoreName)) continue

            val className = CaseFormat.LOWER_UNDERSCORE.to(CaseFormat.UPPER_CAMEL, lowerUnderscoreName) + "Block"

            val filePathRelativeToSourceRoot = "./${BLOCK_PACKAGE.replace(".", "/")}/$className.kt"
            if (sourcesDir.resolve(filePathRelativeToSourceRoot).exists()) continue

            val materialArgs = arrayOf(
                block.get("id").toInt(),
                "minecraft:$lowerUnderscoreName",
                block.get("minStateId").toInt(),
                block.get("hardness").toFloat(),
                block.get("transparent").toBoolean(),
                block.get("filterLight").toInt(),
                block.get("emitLight").toInt(),
                collisionShapes.getValue(lowerUnderscoreName).toInt()
            )

            val type = TypeSpec.classBuilder(className)
                .apply {
                    if (block.get("states").asList().isNotEmpty()) {
                        addModifiers(KModifier.DATA)
                        primaryConstructor(
                            FunSpec.constructorBuilder().addParameter(
                                ParameterSpec.builder("PLACEHOLDER", Unit::class).defaultValue("Unit").build()
                            ).build()
                        )
                        addProperty(PropertySpec.builder("PLACEHOLDER", Unit::class).initializer("PLACEHOLDER").build())
                    }
                }
                .superclass(BLOCK_TYPE)
                .addType(
                    TypeSpec.companionObjectBuilder()
                        .addSuperinterface(
                            MATERIAL_TYPE.parameterizedBy(ClassName(BLOCK_PACKAGE, className)),
                            CodeBlock.of(
                                "material(\n${materialLines.joinToString(",\n") { "    $it" }}\n)",
                                *materialArgs
                            )
                        )
                        .build()
                )
                .build()

            FileSpec.builder(BLOCK_PACKAGE, className)
                .addType(type)
                .build()
                .writeTo(outputDir)
        }
    }

    private fun generateBlocksList(blocks: List<JsonAny>) {
        val names = blocks
            .map { it.get("name").toString() }
            .filter { !BLOCK_BLACKLIST.contains(it) }
            .map { CaseFormat.LOWER_UNDERSCORE.to(CaseFormat.UPPER_CAMEL, it) + "Block" }

        val property = PropertySpec.builder(
            "GENERATED_BLOCKS",
            List::class.asTypeName().parameterizedBy(MATERIAL_TYPE.parameterizedBy(STAR))
        )
            .initializer("listOf(\n${names.joinToString(",\n")}\n)")
            .build()

        FileSpec.builder(BLOCK_PACKAGE, "Blocks")
            .addProperty(property)
            .build()
            .writeTo(outputDir)
    }
}
