package space.uranos.mdsp.generator

import com.google.common.base.CaseFormat
import com.jsoniter.JsonIterator
import com.squareup.kotlinpoet.*
import com.squareup.kotlinpoet.ParameterizedTypeName.Companion.parameterizedBy
import space.uranos.mdsp.JsonAny
import java.io.File
import kotlin.reflect.KClass

class EntitiesGenerator(
    private val workingDir: File,
    private val outputDir: File,
    private val sourcesDir: File
) {
    fun generate() {
        val entitiesJson = workingDir.resolve("entities.json").readText()
        val types = JsonIterator.deserialize(entitiesJson).asList()

        generateEntityTypes(types)
        generateEntityStubs(types)
        generateEntityTypeList(types)
    }

    private fun generateEntityTypes(types: List<JsonAny>) {
        for (entity in types) {
            val entityClassName =
                CaseFormat.LOWER_UNDERSCORE.to(CaseFormat.UPPER_CAMEL, entity.get("name").toString())!! +
                    "Entity"

            val entityClass = ClassName(ENTITY_PACKAGE, entityClassName)

            val name = entityClassName + "Type"

            val type = TypeSpec.classBuilder(name)
                .addModifiers(KModifier.ABSTRACT)
                .primaryConstructor(FunSpec.constructorBuilder().addModifiers(KModifier.INTERNAL).build())
                .addSuperinterface(ENTITY_TYPE_TYPE.parameterizedBy(entityClass))
                .addProperty(
                    PropertySpec
                        .builder(
                            "interfaceType",
                            KClass::class.asTypeName().parameterizedBy(entityClass),
                            KModifier.OVERRIDE
                        )
                        .initializer("%T::class", entityClass)
                        .build()
                )
                .addProperty(
                    PropertySpec
                        .builder("numericID", Int::class, KModifier.OVERRIDE)
                        .initializer("%L", entity.get("id").toInt())
                        .build()
                )
                .addProperty(
                    PropertySpec
                        .builder("id", String::class, KModifier.OVERRIDE)
                        .initializer("%S", "minecraft:" + entity.get("name").toString())
                        .build()
                )
                .addProperty(
                    PropertySpec
                        .builder("width", Float::class, KModifier.OVERRIDE)
                        .initializer("%Lf", entity.get("width").toFloat())
                        .build()
                )
                .addProperty(
                    PropertySpec
                        .builder("height", Float::class, KModifier.OVERRIDE)
                        .initializer("%Lf", entity.get("height").toFloat())
                        .build()
                )
                .build()

            FileSpec.builder(ENTITY_PACKAGE, name)
                .addType(type)
                .build()
                .writeTo(outputDir)
        }
    }

    private fun generateEntityStubs(types: List<JsonAny>) {
        for (entity in types) {
            val name = CaseFormat.LOWER_UNDERSCORE.to(CaseFormat.UPPER_CAMEL, entity.get("name").toString())!! +
                "Entity"

            if (name == "PaintingEntity") continue

            val path = "./${ENTITY_PACKAGE.replace(".", "/")}/$name.kt"
            if (sourcesDir.resolve(path).exists()) continue

            val type = TypeSpec.interfaceBuilder(name)
                .addSuperinterface(ENTITY_TYPE)
                .addType(
                    TypeSpec.companionObjectBuilder("Type")
                        .superclass(ClassName(ENTITY_PACKAGE, name + "Type"))
                        .build()
                )
                .build()

            FileSpec.builder(ENTITY_PACKAGE, name)
                .addType(type)
                .build()
                .writeTo(outputDir)
        }
    }

    private fun generateEntityTypeList(entities: List<JsonAny>) {
        val names = entities.asSequence()
            .map { it.get("name").toString() }
            .map { CaseFormat.LOWER_UNDERSCORE.to(CaseFormat.UPPER_CAMEL, it) + "Entity" }

        val listProperty = PropertySpec.builder(
            "ENTITY_TYPES",
            List::class.asTypeName().parameterizedBy(ENTITY_TYPE_TYPE.parameterizedBy(STAR))
        )
            .initializer("listOf(\n${names.joinToString(",\n")}\n)")
            .build()

        val typeVariableName = TypeVariableName("T", ENTITY_TYPE)

        val typeProperty = PropertySpec.builder("type", ENTITY_TYPE_TYPE.parameterizedBy(typeVariableName))
            .addTypeVariable(typeVariableName)
            .receiver(typeVariableName)
            .addAnnotation(AnnotationSpec.builder(Suppress::class).addMember(""""UNCHECKED_CAST"""").build())
            .getter(
                FunSpec.getterBuilder()
                    .addStatement(
                        "return when(this) {\n${names.joinToString("\n") { "is $it -> $it" }.prependIndent("    ")}\n" +
                            "    else -> error(\"Unknown entity type\")\n} as EntityType<T>"
                    )
                    .build()
            )
            .build()

        FileSpec.builder(ENTITY_PACKAGE, "EntityTypes")
            .addProperty(typeProperty)
            .addProperty(listProperty)
            .build()
            .writeTo(outputDir)
    }
}
