package space.uranos

import kotlinx.coroutines.NonCancellable
import kotlinx.coroutines.withContext

interface Scheduler {
    /**
     * Executes [block] after [delay] ticks passed.
     *
     * If you need the return value of [block], use [runAfter].
     */
    fun executeAfter(delay: Long, block: suspend () -> Unit): Task

    /**
     * Executes [block] every [interval] ticks.
     *
     * @param delay The ticks to pass before the first execution.
     */
    fun executeRepeating(interval: Long, delay: Long = interval, block: suspend () -> Unit): Task

    /**
     * Executes [block] when the scheduler (and therefore the server) is shutdown.
     */
    fun registerShutdownTask(block: suspend () -> Unit): Task

    /**
     * Runs [block] after [delay] ticks passed.
     *
     * The task is cancelled when the current coroutine scope is cancelled.
     *
     * If you do not need the return value of [block], you should use [executeAfter] instead.
     */
    suspend fun <R> runAfter(delay: Long, block: suspend () -> R): R

    /**
     * Like [runAfter], but the task is *not* cancelled when the current coroutine scope is cancelled.
     */
    suspend fun <R> runDetachedAfter(delay: Long, block: suspend () -> R): R =
        withContext(NonCancellable) { runAfter(delay, block) }

    interface Task {
        fun cancel()
    }
}
