package space.uranos

import space.uranos.entity.Position
import space.uranos.util.numbers.clamp
import space.uranos.world.VoxelLocation
import kotlin.math.abs
import kotlin.math.pow
import kotlin.math.roundToInt
import kotlin.math.sqrt

data class Vector(val x: Double, val y: Double, val z: Double) {
    /**
     * Returns a VoxelLocation by converting [x], [y] and z to Int and clamping y between 0 and 255.
     */
    fun toVoxelLocation() = VoxelLocation(x.toInt(), y.toInt().clamp(0..255).toUByte(), z.toInt())
    fun roundToBlock() = VoxelLocation(x.roundToInt(), y.toInt().clamp(0..255).toUByte(), z.roundToInt())
    fun asPosition() = Position(x, y, z)

    operator fun plus(other: Vector) = Vector(x + other.x, y + other.y, z + other.z)
    operator fun minus(other: Vector) = Vector(x - other.x, y - other.y, z - other.z)
    operator fun div(other: Vector) = Vector(x / other.x, y / other.y, z / other.z)
    operator fun times(other: Vector) = Vector(x * other.x, y * other.y, z * other.z)

    fun plus(x: Double, y: Double, z: Double) = Vector(this.x + x, this.y + y, this.z + z)
    fun minus(x: Double, y: Double, z: Double) = Vector(this.x - x, this.y - y, this.z - z)
    fun div(x: Double, y: Double, z: Double) = Vector(this.x / x, this.y / y, this.z / z)
    fun times(x: Double, y: Double, z: Double) = Vector(this.x * x, this.y * y, this.z * z)

    operator fun get(part: CoordinatePart): Double = when (part) {
        CoordinatePart.X -> x
        CoordinatePart.Y -> y
        CoordinatePart.Z -> z
    }

    companion object {
        /**
         * Returns a Vector with the maximum x, y and z values of [a] and [b].
         */
        fun withHigherValuesOf(a: Vector, b: Vector) = Vector(maxOf(a.x, b.x), maxOf(a.y, b.y), maxOf(a.z, b.z))

        /**
         * Returns a Vector with the minimum x, y and z values of [a] and [b].
         */
        fun withLowerValuesOf(a: Vector, b: Vector) = Vector(minOf(a.x, b.x), minOf(a.y, b.y), minOf(a.z, b.z))

        /**
         * Returns the (euclidean) distance between [a] and [b].
         */
        fun distanceBetween(a: Vector, b: Vector): Double {
            val (x, y, z) = abs(a - b)

            return sqrt(x.pow(2) + y.pow(2) + z.pow(2))
        }

        val ZERO = Vector(0.0, 0.0, 0.0)
    }
}

/**
 * Returns a Vector with the absolute values of x, y and z.
 */
fun abs(vector: Vector) = Vector(abs(vector.x), abs(vector.y), abs(vector.z))
