package space.uranos.chat

import com.squareup.moshi.*
import space.uranos.util.moshi

sealed class ChatComponent {
    abstract val bold: Boolean
    abstract val italic: Boolean
    abstract val underlined: Boolean
    abstract val strikethrough: Boolean
    abstract val obfuscated: Boolean
    abstract val color: ChatColor?
    abstract val extra: List<ChatComponent>

    object JSONAdapter {
        @FromJson
        fun fromJson(@Suppress("UNUSED_PARAMETER") reader: JsonReader): ChatComponent? {
            throw UnsupportedOperationException("ChatComponent cannot be deserialized.")
        }

        @ToJson
        fun toJson(writer: JsonWriter, value: ChatComponent?) {
            @Suppress("UNCHECKED_CAST")
            if (value == null) writer.nullValue()
            else (moshi.adapter(value::class.java) as JsonAdapter<ChatComponent>).toJson(writer, value)
        }
    }

    fun toJson(): String = moshi.adapter(ChatComponent::class.java).toJson(this)
}

data class TextComponent(
    val text: String,
    override val bold: Boolean = false,
    override val italic: Boolean = false,
    override val underlined: Boolean = false,
    override val strikethrough: Boolean = false,
    override val obfuscated: Boolean = false,
    override val color: ChatColor? = null,
    override val extra: List<ChatComponent> = emptyList()
) : ChatComponent() {
    companion object {
        /**
         * Creates a new [TextComponent] instance using [text].
         */
        infix fun of(text: String) = TextComponent(text)
    }

    // We cannot use the automatic generation because extra must be omitted if it is empty (-_-)
    object JSONAdapter {
        @FromJson
        fun fromJson(@Suppress("UNUSED_PARAMETER") reader: JsonReader): TextComponent {
            throw UnsupportedOperationException("TextComponent cannot be deserialized.")
        }

        @ToJson
        fun toJson(writer: JsonWriter, value: TextComponent) {
            writer.beginObject()
            writer.name("text")
            writer.value(value.text)
            writer.name("bold")
            writer.value(value.bold)
            writer.name("italic")
            writer.value(value.italic)
            writer.name("underlined")
            writer.value(value.underlined)
            writer.name("strikethrough")
            writer.value(value.strikethrough)
            writer.name("obfuscated")
            writer.value(value.obfuscated)

            if (value.color != null) {
                writer.name("color")
                writer.value(ChatColor.JSONAdapter.toJson(value.color))
            }

            if (value.extra.isNotEmpty()) {
                writer.name("extra")
                writer.beginArray()
                value.extra.forEach { ChatComponent.JSONAdapter.toJson(writer, it) }
                writer.endArray()
            }

            writer.endObject()
        }
    }
}
