package space.uranos.entity

import com.google.common.cache.CacheBuilder
import com.google.common.cache.CacheLoader
import space.uranos.event.EventBusWrapper
import space.uranos.player.Player
import space.uranos.world.Chunk
import space.uranos.world.World
import java.util.UUID

interface Entity {
    /**
     * The UUID of this entity.
     *
     * If the entity is a player and the server uses authentication, it is the player's UUID, if it is not,
     * it is a UUIDv3 generated from a string consisting of "OfflinePlayer:" and the player's username.
     *
     * Otherwise, it is usually randomly generated.
     */
    val uuid: UUID

    /**
     * An integer unique to this entity which will not be persisted, for example when the entity is serialized.
     */
    val numericID: Int

    /**
     * Players that can see this entity.
     */
    val viewers: MutableSet<Player>

    fun belongsToChunk(key: Chunk.Key): Boolean

    val world: World?
    suspend fun setWorld(world: World?)

    /**
     * If players should be added to [viewers] when they enter [world].
     */
    var visibleToNewPlayers: Boolean
}

/**
 * Returns [Entity.world] if it is not null, otherwise throws [IllegalStateException].
 */
fun Entity.getWorldOrFail() = world ?: throw IllegalStateException("This entity has not been spawned")

private val eventBusWrapperCache = CacheBuilder.newBuilder()
    .weakKeys()
    .build(object : CacheLoader<Entity, EventBusWrapper<Entity>>() {
        override fun load(key: Entity): EventBusWrapper<Entity> = EventBusWrapper(key)
    })

@Suppress("UNCHECKED_CAST")
val <T : Entity> T.events
    get() = eventBusWrapperCache.get(this) as EventBusWrapper<T>
