package space.uranos.event

import kotlinx.coroutines.suspendCancellableCoroutine
import space.uranos.plugin.Plugin
import java.util.concurrent.ConcurrentSkipListSet
import kotlin.coroutines.resume
import kotlin.reflect.KClass

abstract class EventEmitter<E: Event> {
    protected val handlers = ConcurrentSkipListSet<EventHandler<*>>()

    /**
     * Registers a new event handler.
     */
    open fun <T: E> on(
        eventType: KClass<T>,
        handlerPosition: EventHandlerPosition = EventHandlerPosition.NORMAL,
        fn: suspend (event: T) -> Unit
    ): EventHandler<T> {
        val handler = EventHandler(this, eventType, Plugin.getCalling(), handlerPosition, fn)
        handlers.add(handler)
        return handler
    }

    /**
     * Registers a new event handler.
     *
     * The handler function receives a function to remove the event handler as second parameter.
     */
    fun <T: E> onRemovable(
        eventType: KClass<T>,
        handlerPosition: EventHandlerPosition = EventHandlerPosition.NORMAL,
        fn: suspend (event: T, remove: () -> Unit) -> Unit
    ): EventHandler<T> {
        lateinit var handler: EventHandler<T>

        @Suppress("JoinDeclarationAndAssignment") // false positive
        handler = on(eventType, handlerPosition) {
            fn(it, handler::remove)
        }

        return handler
    }

    /**
     * Registers a new event handler which is automatically removed after it was invoked.
     */
    fun <T: E> once(
        eventType: KClass<T>,
        handlerPosition: EventHandlerPosition = EventHandlerPosition.NORMAL,
        fn: suspend (event: T) -> Unit
    ): EventHandler<T> = onRemovable(eventType, handlerPosition) { event, remove -> remove(); fn(event) }

    /**
     * Suspends until an event of type [T] is emitted for which [predicate] returns true and returns it.
     */
    suspend fun <T : E> waitFor(
        eventType: KClass<T>,
        handlerPosition: EventHandlerPosition = EventHandlerPosition.NORMAL,
        predicate: (event: T) -> Boolean = { true }
    ): T =
        suspendCancellableCoroutine { c ->
            lateinit var handler: EventHandler<*>

            handler = EventHandler(this, eventType, Plugin.getCalling(), handlerPosition) {
                if (predicate(it)) {
                    handler.remove()
                    c.resume(it)
                }
            }

            handlers.add(handler)

            c.invokeOnCancellation {
                handlers.remove(handler)
            }
        }

    /**
     * @return Whether the event handler was not already removed.
     */
    fun remove(eventHandler: EventHandler<*>): Boolean = handlers.remove(eventHandler)

    /**
     * Registers a new event handler.
     */
    inline fun <reified T: E> on(
        handlerPosition: EventHandlerPosition = EventHandlerPosition.NORMAL,
        noinline fn: suspend (event: T) -> Unit
    ): EventHandler<T> = on(T::class, handlerPosition, fn)

    /**
     * Registers a new event handler.
     *
     * The handler function receives a function to remove the event handler as second parameter.
     */
    inline fun <reified T: E> onRemovable(
        handlerPosition: EventHandlerPosition = EventHandlerPosition.NORMAL,
        noinline fn: suspend (event: T, remove: () -> Unit) -> Unit
    ): EventHandler<T> = onRemovable(T::class, handlerPosition, fn)

    /**
     * Registers a new event handler which is automatically removed after it was invoked.
     */
    inline fun <reified T: E> once(
        handlerPosition: EventHandlerPosition = EventHandlerPosition.NORMAL,
        noinline fn: suspend (event: T) -> Unit
    ): EventHandler<T> = once(T::class, handlerPosition, fn)

    /**
     * Suspends until an event of type [T] is emitted for which [predicate] returns true and returns it.
     */
    suspend inline fun <reified T : E> waitFor(
        position: EventHandlerPosition = EventHandlerPosition.NORMAL,
        noinline predicate: (event: T) -> Boolean = { true }
    ): T =
        waitFor(T::class, position, predicate)
}

abstract class EventBus: EventEmitter<Event>() {
    /**
     * Invokes all previously registered event handlers sorted by their handlerPosition.
     *
     * The coroutine context is inherited.
     *
     * @return [event]
     */
    abstract suspend fun <T : Event> emit(event: T): T
}
