package space.uranos.net.packet

import com.google.common.cache.CacheBuilder
import com.google.common.cache.CacheLoader
import com.google.common.cache.LoadingCache
import io.netty.buffer.ByteBuf
import io.netty.buffer.Unpooled
import java.time.Duration
import kotlin.reflect.KClass

abstract class PacketCodec<T : Packet>(val id: Int, val dataType: KClass<T>)

abstract class IncomingPacketCodec<T : IncomingPacket>(id: Int, dataType: KClass<T>) : PacketCodec<T>(id, dataType) {
    abstract fun decode(msg: ByteBuf): T
}

abstract class OutgoingPacketCodec<T : OutgoingPacket>(
    id: Int,
    dataType: KClass<T>,
    cacheOptions: CacheOptions? = null
) : PacketCodec<T>(id, dataType) {
    protected abstract fun T.encode(dst: ByteBuf)

    private val cache: LoadingCache<T, ByteBuf>? = cacheOptions?.let { options ->
        CacheBuilder.newBuilder()
            .weakKeys()
            .maximumSize(options.limit)
            .expireAfterAccess(options.expirationDuration)
            .concurrencyLevel(2)
            .removalListener<T, ByteBuf> { it.value.release() }
            .build(object : CacheLoader<T, ByteBuf>() {
                override fun load(packet: T): ByteBuf {
                    val dst = Unpooled.buffer()
                    packet.encode(dst)
                    return dst
                }
            })
    }

    @JvmName("encodePacket")
    fun encode(packet: T, dst: ByteBuf) {
        if (cache == null) packet.encode(dst)
        else dst.writeBytes(cache.get(packet))
    }

    data class CacheOptions(
        val limit: Long,
        val expirationDuration: Duration
    )
}
