package space.uranos.server

import kotlinx.coroutines.CoroutineDispatcher
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.SupervisorJob
import space.uranos.Registry
import space.uranos.Scheduler
import space.uranos.command.Command
import space.uranos.entity.Entity
import space.uranos.entity.EntityType
import space.uranos.event.EventBus
import space.uranos.event.EventHandlerPositionManager
import space.uranos.logging.Logger
import space.uranos.logging.LoggingOutputProvider
import space.uranos.net.Session
import space.uranos.player.Player
import space.uranos.plugin.PluginManager
import space.uranos.recipe.Recipe
import space.uranos.util.newSingleThreadDispatcher
import space.uranos.world.BiomeRegistry
import space.uranos.world.Dimension
import space.uranos.world.WorldRegistry
import java.io.File

abstract class Server {
    abstract val eventBus: EventBus
    abstract val eventHandlerPositions: EventHandlerPositionManager

    /**
     * A coroutine dispatcher that is confined to the server thread.
     */
    val dispatcher: CoroutineDispatcher = newSingleThreadDispatcher("Server")

    /**
     * A coroutine scope that is cancelled when the server is shutdown.
     */
    val scope = CoroutineScope(SupervisorJob() + dispatcher)

    /**
     * All sessions connected to the server.
     */
    abstract val sessions: Collection<Session>

    /**
     * All players connected to the server.
     */
    abstract val players: Collection<Player>

    abstract val pluginManager: PluginManager
    abstract val serverDirectory: File
    val cacheDirectory: File get() = serverDirectory.resolve("cache")

    abstract val worldRegistry: WorldRegistry
    val dimensionRegistry: Registry<Dimension> = Registry()
    val recipeRegistry: Registry<Recipe> = Registry()
    val commandRegistry: Registry<Command> = Registry()
    val biomeRegistry: BiomeRegistry = BiomeRegistry()

    abstract val loggingOutputProvider: LoggingOutputProvider
    abstract val scheduler: Scheduler

    abstract val developmentMode: Boolean
    abstract val minimumLogLevel: Logger.Level

    /**
     * Set of all existing [Entity] instances.
     */
    abstract val entities: Set<Entity>

    /**
     * Create a new entity of [type].
     *
     * @throws IllegalArgumentException When [type] is [PlayerEntity][space.uranos.entity.PlayerEntity].
     */
    abstract fun <T : Entity> create(type: EntityType<T>): T
    inline fun <reified T : Entity> create(): T = create(EntityType.byInterfaceType(T::class))

    /**
     * Initiates shutting down the server.
     */
    abstract fun shutdown()

    companion object {
        const val TICKS_PER_SECOND = 20
    }
}
