package space.uranos.util

import kotlinx.coroutines.coroutineScope
import kotlinx.coroutines.launch
import java.lang.ref.WeakReference
import kotlin.reflect.KProperty

class TickSynchronizationContainer {
    // If it is too resource-intensive to create a new object for every property, one instance could be used per TickSynchronizationContainer
    private val delegates = mutableSetOf<Delegate<*>>()

    suspend fun tick() {
        coroutineScope {
            delegates.forEach { launch { it.tick() } }
        }
    }

    fun <T> ifChanged(
        initialValue: T,
        onChange: ((value: T) -> Unit)? = null,
        onTick: suspend (value: T) -> Unit
    ): Delegate<T> =
        Delegate(initialValue, onChange, { value, changed -> if (changed) onTick(value) }).also { delegates.add(it) }

    operator fun <T> invoke(
        initialValue: T,
        onChange: ((value: T) -> Unit)? = null,
        onTick: suspend (value: T, changed: Boolean) -> Unit
    ): Delegate<T> = Delegate(initialValue, onChange, onTick).also { delegates.add(it) }

    inner class Delegate<T>(
        initialValue: T,
        private val onChange: ((value: T) -> Unit)?,
        private val onTick: suspend (value: T, changed: Boolean) -> Unit
    ) {
        operator fun getValue(thisRef: Any?, property: KProperty<*>): T = value

        operator fun setValue(thisRef: Any?, property: KProperty<*>, value: T) {
            val oldValue = oldValueRef.get()
            if (oldValue == value) return
            onChange?.invoke(value)
            this.value = value
            changed = true
        }

        private var oldValueRef = WeakReference(initialValue)
        private var changed = true

        var value: T = initialValue

        suspend fun tick() {
            onTick(value, changed)

            oldValueRef = WeakReference(value)
            changed = false
        }
    }
}
