package space.uranos.util.collections

import java.util.Spliterator
import java.util.function.Predicate
import java.util.stream.Stream

abstract class WatchableSet<T>(private val backingSet: MutableSet<T>) : MutableSet<T> by backingSet {
    abstract fun onAdd(element: T)
    abstract fun onRemove(element: T)
    abstract fun beforeAdd(element: T)

    override fun add(element: T): Boolean {
        beforeAdd(element)
        val added = backingSet.add(element)
        if (added) onAdd(element)
        return added
    }

    override fun remove(element: T): Boolean {
        val removed = backingSet.remove(element)
        if (removed) onRemove(element)
        return removed
    }

    override fun addAll(elements: Collection<T>): Boolean = elements.any { add(it) }

    override fun clear() {
        backingSet.forEach { remove(it) }
    }

    override fun removeAll(elements: Collection<T>): Boolean =
        if (elements.size > backingSet.size) backingSet.toSet().any { if (elements.contains(it)) remove(it) else false }
        else elements.any { remove(it) }

    override fun retainAll(elements: Collection<T>): Boolean =
        if (elements.size > backingSet.size) backingSet.toSet().any { if (elements.contains(it)) remove(it) else false }
        else elements.any { remove(it) }

    override fun iterator(): MutableIterator<T> {
        val iterator = backingSet.iterator()
        var current: T? = null

        return object : MutableIterator<T> {
            override fun hasNext(): Boolean = iterator.hasNext()
            override fun next(): T = iterator.next().also { current = it }
            override fun remove() {
                iterator.remove()
                current?.let { onRemove(it) }
            }
        }
    }

    override fun removeIf(filter: Predicate<in T>): Boolean {
        val iterator = iterator()
        var modified = false

        for (item in iterator) {
            if (filter.test(item)) {
                modified = true
                iterator.remove()
            }
        }

        return modified
    }

    override fun parallelStream(): Stream<T> = throw UnsupportedOperationException("Not implemented for WatchableSet")
    override fun stream(): Stream<T> = throw UnsupportedOperationException("Not implemented for WatchableSet")
    override fun spliterator(): Spliterator<T> = throw UnsupportedOperationException("Not implemented for WatchableSet")
}
