package space.uranos.world

import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.SupervisorJob
import space.uranos.Uranos
import space.uranos.entity.Position
import space.uranos.player.Player
import kotlin.math.floor

abstract class Chunk(
    val world: World,
    val key: Key
) {
    private val identifier = "Chunk(${key.x}-${key.z})"

    data class Key(val x: Int, val z: Int) {
        companion object {
            /**
             * Returns the key for the chunk containing the voxel at [location].
             */
            fun from(location: VoxelLocation) = Key(
                floor(location.x.toFloat() / LENGTH).toInt(),
                floor(location.z.toFloat() / LENGTH).toInt()
            )

            fun from(location: Position) = Key(
                floor(location.x.toFloat() / LENGTH).toInt(),
                floor(location.z.toFloat() / LENGTH).toInt()
            )
        }

        fun translateWorldToChunk(x: Int, z: Int) = Pair(Math.floorMod(x, LENGTH), Math.floorMod(z, LENGTH))
        fun translateChunkToWorld(x: Int, z: Int) =
            Pair(this.x * LENGTH + x, this.z * LENGTH + z)

        fun translateWorldToChunk(location: VoxelLocation) = translateWorldToChunk(location.x, location.z)
            .let { (x, z) -> VoxelLocation(x, location.y, z) }

        fun translateChunkToWorld(location: VoxelLocation) = translateChunkToWorld(location.x, location.z)
            .let { (x, z) -> VoxelLocation(x, location.y, z) }
    }

    /**
     * A coroutine scope confined to the world's dispatcher that is cancelled when the chunk is unloaded.
     */
    val scope = CoroutineScope(SupervisorJob() + world.dispatcher)

    /**
     * A list of all players who have locked this chunk.
     *
     * @see [Player.currentlyViewedChunks]
     */
    val viewers get() = Uranos.players.filter { it.currentlyViewedChunks.contains(this) }

    /**
     * Returns the voxel for the specified coordinates **relative to this chunk**.
     *
     * @throws NoSuchElementException If [location] is not in this chunk.
     */
    abstract fun getVoxelRelativeToChunk(location: VoxelLocation): Voxel

    /**
     * Returns the voxel at [location].
     *
     * @throws NoSuchElementException If [location] is not in this chunk.
     */
    abstract fun getVoxel(location: VoxelLocation): Voxel

    /**
     * Alias for [getVoxel].
     */
    operator fun get(location: VoxelLocation) = getVoxel(location)

    /**
     * Returns whether this chunk contains [location].
     */
    operator fun contains(location: VoxelLocation) = Key.from(location) == key

    /**
     * Returns whether this chunk contains [voxel].
     */
    // It does not just compare voxel.chunk with this because voxel.chunk may be expensive to create
    operator fun contains(voxel: Voxel) = voxel.world == world && contains(voxel.location)

    abstract fun getData(player: Player?): ChunkData
    abstract fun getLightData(player: Player?): ChunkLightData

    /**
     * Sets the biome at the specified position. Biome areas are 4x4x4, so [x] and [z] must be in 0..3 and [y] in 0..63.
     */
    abstract fun setBiome(x: Int, y: Int, z: Int, biome: Biome)

    /**
     * Starts loading this chunk into memory.
     *
     * Does nothing if the chunk is already loaded or loading.
     */
    open fun startLoading() {}
    open suspend fun waitUntilLoaded() {}

    open val loaded = true

    companion object {
        const val LENGTH: Int = 16
        const val AREA = LENGTH * LENGTH
        const val SECTION_HEIGHT: Int = 16
        const val BLOCKS_IN_A_SECTION = AREA * SECTION_HEIGHT
    }

    interface Unloadable {
        suspend fun unload()
    }
}
