package space.uranos.world

import space.uranos.world.block.Block

data class ChunkData(
    /**
     * The outer array must have 16 items representing the chunk sections.
     *
     * The inner arrays contain the blocks (x, z, y) of a chunk section and must have a size of 4096.
     * If a chunk section only contains air, it should be null (instead of an array of AirBlock).
     */
    val sections: Array<Array<Block>?>,

    /**
     * 1024 biome IDs (one for each 4x4x4 area), ordered in this order: X, Z, Y.
     */
    val biomes: Array<Biome>,

    /**
     * An array (first x, then z) containing the y coordinates of the highest **solid** block in every column.
     * Will be computed using [sections] if null.
     */
    val solidBlocksHeightmap: ByteArray? = null,

    /**
     * An array (first x, then z) containing the y coordinates of the highest block in every column which is
     * neither [AirBlock][space.uranos.world.block.AirBlock] nor [CaveAirBlock][space.uranos.world.block.CaveAirBlock].
     * Will be computed using [sections] if null.
     */
    val nonAirBlocksHeightmap: ByteArray? = null
) {

    companion object {
        const val BIOME_AREAS_IN_CHUNK = 1024
    }

    override fun equals(other: Any?): Boolean {
        if (this === other) return true
        if (javaClass != other?.javaClass) return false

        other as ChunkData

        if (!sections.contentDeepEquals(other.sections)) return false
        if (!biomes.contentEquals(other.biomes)) return false
        if (solidBlocksHeightmap != null) {
            if (other.solidBlocksHeightmap == null) return false
            if (!solidBlocksHeightmap.contentEquals(other.solidBlocksHeightmap)) return false
        } else if (other.solidBlocksHeightmap != null) return false
        if (nonAirBlocksHeightmap != null) {
            if (other.nonAirBlocksHeightmap == null) return false
            if (!nonAirBlocksHeightmap.contentEquals(other.nonAirBlocksHeightmap)) return false
        } else if (other.nonAirBlocksHeightmap != null) return false

        return true
    }

    override fun hashCode(): Int {
        var result = sections.contentDeepHashCode()
        result = 31 * result + biomes.contentHashCode()
        result = 31 * result + (solidBlocksHeightmap?.contentHashCode() ?: 0)
        result = 31 * result + (nonAirBlocksHeightmap?.contentHashCode() ?: 0)
        return result
    }

    override fun toString(): String = "ChunkData([omitted])"
}
