package space.uranos.world

import space.uranos.CoordinatePart
import space.uranos.Vector
import space.uranos.entity.Position

/**
 * A location consisting of an x, y and z coordinate.
 *
 * This class does not contain any functions for calculations, but [Vector] does.
 * You can convert a VoxelLocation to a Vector using [asVector].
 */
data class VoxelLocation(val x: Int, val y: UByte, val z: Int) {
    /**
     * Converts this VoxelLocation to a Location.
     */
    fun asPosition(): Position = Position(x.toDouble(), y.toDouble(), z.toDouble())

    fun asVector(): Vector = Vector(x.toDouble(), y.toDouble(), z.toDouble())

    /**
     * Converts this VoxelLocation to a Location **and then adds 0.5 to x, y and z**.
     *
     * Example: `VoxelLocation(x = 1, y = 2, z = 3)` becomes `Location(x = 1.5, y = 2.5, z = 3.5)`.
     */
    fun atCenter(): Position = Position(x.toDouble() + 0.5, y.toDouble() + 0.5, z.toDouble() + 0.5)

    /**
     * Converts this VoxelLocation to a Location **and then adds 0.5 to x and z, but not y**.
     *
     * Example: `VoxelLocation(x = 1, y = 2, z = 3)` becomes `Location(x = 1.5, y = 2, z = 3.5)`.
     */
    fun atTopCenter(): Position = Position(x.toDouble() + 0.5, y.toDouble(), z.toDouble() + 0.5)

    /**
     * Returns a VoxelLocation with the maximum x, y and z values of this and [other].
     */
    fun withHighestValues(other: VoxelLocation) = VoxelLocation(maxOf(x, other.x), maxOf(y, other.y), maxOf(z, other.z))

    /**
     * Returns a VoxelLocation with the minimum x, y and z values of this and [other].
     */
    fun withLowestValues(other: VoxelLocation) = VoxelLocation(minOf(x, other.x), minOf(y, other.y), minOf(z, other.z))

    operator fun get(part: CoordinatePart): Int = when (part) {
        CoordinatePart.X -> x
        CoordinatePart.Y -> y.toInt()
        CoordinatePart.Z -> z
    }

    companion object {
        val ZERO = VoxelLocation(0, 0u, 0)

        fun of(x: Int, y: Int, z: Int): VoxelLocation {
            if (y !in 0..255) throw IllegalArgumentException("y must be in 0..255")
            return VoxelLocation(x, y.toUByte(), z)
        }
    }
}




