package space.uranos.world

import kotlinx.coroutines.CoroutineDispatcher
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.withContext
import space.uranos.Vector
import space.uranos.entity.Entity
import space.uranos.util.numbers.untilPossiblyLower

interface World {
    val dispatcher: CoroutineDispatcher
    val dimension: Dimension
    val isFlat: Boolean
    val entities: Collection<Entity>
    val loadedChunks: Map<Chunk.Key, Chunk>

    interface Internals {
        val world: World

        fun addEntity(entity: Entity)
        fun removeEntity(entity: Entity)
    }

    /**
     * This can be any value.
     *
     * Only the first 8 bytes of the SHA-256 hash of this value are sent to the client
     * (and it seems not to use it in any way).
     * A random value will be used if this is null.
     */
    val seed: Long?

    fun getChunk(key: Chunk.Key): Chunk
}

/**
 * Returns the chunk containing the voxel at [location].
 */
fun World.getChunk(location: VoxelLocation): Chunk = getChunk(Chunk.Key.from(location))

/**
 * Returns the voxel at [location].
 */
fun World.getVoxel(location: VoxelLocation): Voxel = getChunk(location).getVoxel(location)

/**
 * Returns all voxels in the cuboid with the corner points [cornerA] and [cornerB].
 *
 * @param hollow Whether the cuboid is hollow
 */
fun World.getVoxelsInCuboid(cornerA: VoxelLocation, cornerB: VoxelLocation, hollow: Boolean = false): List<Voxel> {
    fun getList(a: Int, b: Int) =
        if (hollow) listOf(a, b).distinct()
        else (a untilPossiblyLower b).toList()

    return buildList {
        for (x in getList(cornerA.x, cornerB.x)) {
            for (y in getList(cornerA.y.toInt(), cornerB.y.toInt())) {
                for (z in getList(cornerA.z, cornerB.z)) {
                    add(getVoxel(VoxelLocation(x, y.toUByte(), z)))
                }
            }
        }
    }
}


/**
 * Returns all voxels in a sphere with the specified [center] and [radius].
 */
fun World.getVoxelsInSphere(
    center: VoxelLocation,
    radius: Double,
    wallWidth: Int = Int.MAX_VALUE
): List<Voxel> {
    val centerVector = center.asVector()
    return getVoxelsInCuboid(
        centerVector.plus(radius, radius, radius).toVoxelLocation(),
        centerVector.minus(radius, radius, radius).toVoxelLocation()
    ).filter {
        val distance = Vector.distanceBetween(it.location.asVector(), centerVector)
        distance < radius && distance > radius - 0.5 - wallWidth
    }
}

suspend inline fun <T> runIn(world: World, noinline block: suspend CoroutineScope.() -> T): T =
    withContext(world.dispatcher, block)
