package space.uranos.world.block

import space.uranos.util.KPropertyValuePair
import kotlin.reflect.KClass
import kotlin.reflect.KProperty
import kotlin.reflect.KProperty1
import kotlin.reflect.full.declaredMemberProperties
import kotlin.reflect.full.findAnnotation
import kotlin.reflect.full.isSubtypeOf
import kotlin.reflect.full.starProjectedType
import kotlin.reflect.jvm.jvmErasure

class BlockCodec<T : Block> internal constructor(
    blockClass: KClass<T>,
    val id: Int,
    val firstStateID: Int
) {
    val states: List<Array<KPropertyValuePair<*>>> =
        generateStates(blockClass.declaredMemberProperties.mapNotNull { property ->
            val annotation = property.findAnnotation<Attribute>() ?: return@mapNotNull null
            val type = property.returnType

            @Suppress("UNCHECKED_CAST")
            when {
                type == Int::class.starProjectedType -> BlockAttribute.Int(
                    (property as KProperty1<Block, Int>),
                    annotation.max
                )
                type == Boolean::class.starProjectedType -> BlockAttribute.Boolean(property as KProperty1<Block, Boolean>)
                type.isSubtypeOf(Enum::class.starProjectedType) -> BlockAttribute.Enum(property as KProperty1<Block, Enum<*>>)
                else -> throw IllegalAttributeTargetType()
            }
        })

    val lastStateID get() = firstStateID + states.size

    class IllegalAttributeTargetType : Exception("The type of the target property is not allowed for attributes")

    fun getStateID(block: T): Int {
        if (states.isEmpty()) return firstStateID

        val values = mutableMapOf<KProperty<*>, Any>()

        val index = states.indexOfFirst { propertyValuePairs ->
            propertyValuePairs.all {
                values.getOrPut(it.property) { it.property.call(block)!! } == it.value
            }
        }

        // If this happens, we did something wrong with adding the attributes or with clamping integer attributes
        if (index == -1) throw Error("The state of block (${block.material().id}) has no ID")

        return firstStateID + index
    }

    private sealed class BlockAttribute<T : Any> {
        abstract val property: KProperty1<Block, T>

        data class Boolean(override val property: KProperty1<Block, kotlin.Boolean>) : BlockAttribute<kotlin.Boolean>()
        data class Int(override val property: KProperty1<Block, kotlin.Int>, val max: kotlin.Int) :
            BlockAttribute<kotlin.Int>()

        data class Enum(override val property: KProperty1<Block, kotlin.Enum<*>>) : BlockAttribute<kotlin.Enum<*>>()
    }

    companion object {
        private fun generateStates(properties: List<BlockAttribute<*>>): List<Array<KPropertyValuePair<*>>> {
            if (properties.isEmpty()) return emptyList()

            val current = properties[0]

            val allowedValues: Array<out Any> = when (current) {
                is BlockAttribute.Boolean -> arrayOf(true, false)
                is BlockAttribute.Int -> (0..current.max).toList().toTypedArray()
                is BlockAttribute.Enum -> current.property.returnType.jvmErasure.java.enumConstants
            }

            val rest = generateStates(properties.drop(1))
            return allowedValues.flatMap { value ->
                if (rest.isEmpty()) listOf(arrayOf(KPropertyValuePair(current.property, value)))
                else rest.map { arrayOf(KPropertyValuePair(current.property, value), *it) }
            }
        }
    }
}
