package space.uranos.nbt

import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.IOException
import kotlin.reflect.KClass

internal fun stringIf(condition: Boolean, string: String = " ") = if (condition) string else ""

abstract class NBTType<T : Any> internal constructor(val typeClass: KClass<*>, val id: Byte) {
    abstract fun toSNBT(value: T, pretty: Boolean = false): String
    abstract fun readValue(source: DataInputStream): T
    abstract fun writeValue(destination: DataOutputStream, value: T)

    fun writeIDAndValue(destination: DataOutputStream, value: T) {
        destination.writeByte(id.toInt())
        writeValue(destination, value)
    }

    fun writeNamedTag(destination: DataOutputStream, name: String, value: T) {
        destination.writeByte(id.toInt())
        destination.writeUTF(name)
        writeValue(destination, value)
    }

    companion object {
        const val END_TAG_ID: Byte = 0

        val all by lazy {
            setOf<NBTType<*>>(
                NBTByte,
                NBTShort,
                NBTInt,
                NBTLong,
                NBTFloat,
                NBTDouble,
                NBTByteArray,
                NBTString,
                NBTList,
                NBTCompound,
                NBTLongArray
            )
        }

        val byID: Map<Byte, NBTType<*>> by lazy { all.map { it.id to it }.toMap() }

        @Suppress("UNCHECKED_CAST")
        fun <T : Any> of(value: T): NBTType<T>? = all.find { it.typeClass.isInstance(value) } as NBTType<T>?
    }
}

object NBTByte : NBTType<Byte>(Byte::class, 1) {
    override fun toSNBT(value: Byte, pretty: Boolean): String = "${value}b"
    override fun readValue(source: DataInputStream): Byte = source.readByte()
    override fun writeValue(destination: DataOutputStream, value: Byte) = destination.writeByte(value.toInt())
}

object NBTShort : NBTType<Short>(Short::class, 2) {
    override fun toSNBT(value: Short, pretty: Boolean): String = "${value}s"
    override fun readValue(source: DataInputStream): Short = source.readShort()
    override fun writeValue(destination: DataOutputStream, value: Short) = destination.writeShort(value.toInt())
}

object NBTInt : NBTType<Int>(Integer::class, 3) {
    override fun toSNBT(value: Int, pretty: Boolean): String = "$value"
    override fun readValue(source: DataInputStream): Int = source.readInt()
    override fun writeValue(destination: DataOutputStream, value: Int) = destination.writeInt(value)
}

object NBTLong : NBTType<Long>(Long::class, 4) {
    override fun toSNBT(value: Long, pretty: Boolean): String = "${value}L"
    override fun readValue(source: DataInputStream): Long = source.readLong()
    override fun writeValue(destination: DataOutputStream, value: Long) = destination.writeLong(value)
}

object NBTFloat : NBTType<Float>(Float::class, 5) {
    override fun toSNBT(value: Float, pretty: Boolean): String = "${value}f"
    override fun readValue(source: DataInputStream): Float = source.readFloat()
    override fun writeValue(destination: DataOutputStream, value: Float) = destination.writeFloat(value)
}

object NBTDouble : NBTType<Double>(Double::class, 6) {
    override fun toSNBT(value: Double, pretty: Boolean): String = "$value"
    override fun readValue(source: DataInputStream): Double = source.readDouble()
    override fun writeValue(destination: DataOutputStream, value: Double) = destination.writeDouble(value)
}

object NBTByteArray : NBTType<ByteArray>(ByteArray::class, 7) {

    // Not pretty, because no one reads byte arrays
    override fun toSNBT(value: ByteArray, pretty: Boolean): String =
        "[B;" + value.joinToString(",") + "]"

    override fun readValue(source: DataInputStream): ByteArray {
        val length = source.readInt()
        val array = ByteArray(length)
        source.read(array, 0, length)
        return array
    }

    override fun writeValue(destination: DataOutputStream, value: ByteArray) {
        destination.writeInt(value.size)
        destination.write(value)
    }
}

object NBTString : NBTType<String>(String::class, 8) {
    override fun toSNBT(value: String, pretty: Boolean): String =
        "\"${value.replace("\"", "\\\"").replace("\n", "\\n")}\""

    override fun readValue(source: DataInputStream): String = source.readUTF()
    override fun writeValue(destination: DataOutputStream, value: String) = destination.writeUTF(value)
}

object NBTCompound : NBTType<NBT>(NBT::class, 10) {
    override fun toSNBT(value: NBT, pretty: Boolean): String {
        val separator = ",${stringIf(pretty, "\n")}"

        val entries = value.data.entries.joinToString(separator) { (name, v) ->
            "\"$name\"" + ":" + stringIf(pretty) + of(v)!!.toSNBT(v, pretty)
        }.run {
            if (pretty) prependIndent("  ")
            else this
        }

        return "{" + stringIf(pretty, "\n") + entries + stringIf(pretty, "\n") + "}"
    }

    override fun readValue(source: DataInputStream): NBT {
        val data = mutableMapOf<String, Any>()

        while (true) {
            val typeID = source.readByte()
            if (typeID == END_TAG_ID) break

            val type = byID[typeID] ?: throw IOException("The NBT data contains an unknown type ID: $typeID")
            val name = source.readUTF()
            data[name] = type.readValue(source)
        }

        return NBT(data)
    }

    fun readTopLevelValue(source: DataInputStream): NBT {
        val data = mutableMapOf<String, Any>()

        while (source.available() != 0) {
            val typeID = source.readByte()
            if (typeID == END_TAG_ID) throw IOException("TAG_END is not allowed on the top level")

            val type = byID[typeID] ?: throw IOException("The NBT data contains an unknown type ID: $typeID")
            val name = source.readUTF()
            data[name] = type.readValue(source)
        }

        return NBT(data)
    }

    override fun writeValue(destination: DataOutputStream, value: NBT) {
        value.data.forEach { (name, v) ->
            of(v)!!.writeNamedTag(destination, name, v)
        }

        destination.writeByte(END_TAG_ID.toInt())
    }
}

object NBTLongArray : NBTType<LongArray>(LongArray::class, 12) {
    override fun toSNBT(value: LongArray, pretty: Boolean): String =
        "[L;" + value.joinToString(",") + "]"

    override fun readValue(source: DataInputStream): LongArray {
        val length = source.readInt()
        val array = LongArray(length)
        repeat(length) {
            array[it] = source.readLong()
        }

        return array
    }

    override fun writeValue(destination: DataOutputStream, value: LongArray) {
        destination.writeInt(value.size)
        value.forEach { destination.writeLong(it) }
    }
}
