package space.uranos.net

import io.netty.buffer.ByteBuf
import io.netty.buffer.ByteBufOutputStream
import space.uranos.nbt.NBT
import space.uranos.world.VoxelLocation
import java.io.DataOutputStream
import java.io.IOException
import java.util.UUID
import kotlin.experimental.and

/**
 * Extension functions on [ByteBuf] for reading and writing data types used by the Minecraft protocol
 *
 * You can access them using the [with] function (`with(MinecraftDataTypes) { ... }`).
 */
object MinecraftProtocolDataTypes {
    /**
     * Reads a variable length integer.
     *
     * @throws IOException If the VarInt is longer than the maximum of 5 bytes
     * @see <a href="https://wiki.vg/Protocol#VarInt_and_VarLong">https://wiki.vg/Protocol#VarInt_and_VarLong</a>
     */
    fun ByteBuf.readVarInt(): Int {
        var bytesRead = 0
        var result = 0

        do {
            if (bytesRead == 5) throw IOException("VarInt is longer than the maximum of 5 bytes")

            val byte = readByte()
            val value: Int = (byte and 0b01111111).toInt()
            result = result or (value shl (7 * bytesRead))
            bytesRead += 1
        } while ((byte and 0b10000000.toByte()).toInt() != 0)

        return result
    }

    /**
     * Checks if a variable length integer can be read.
     */
    fun ByteBuf.varIntReadable(): Boolean {
        if (readableBytes() > 5) {
            // maximum VarInt size
            return true
        }

        val initialIndex = readerIndex()
        do {
            if (readableBytes() < 1) {
                readerIndex(initialIndex)
                return false
            }

            val value = readByte().toInt()
        } while ((value and 0b10000000) != 0)

        readerIndex(initialIndex)
        return true
    }

    /**
     * Writes a variable length integer.
     *
     * @see <a href="https://wiki.vg/Protocol#VarInt_and_VarLong">https://wiki.vg/Protocol#VarInt_and_VarLong</a>
     */
    fun ByteBuf.writeVarInt(value: Int): ByteBuf {
        var v = value
        var part: Byte
        while (true) {
            part = (v and 0x7F).toByte()
            v = v ushr 7
            if (v != 0) {
                part = (part.toInt() or 0x80).toByte()
            }
            writeByte(part.toInt())
            if (v == 0) {
                break
            }
        }

        return this
    }

    /**
     * Reads a variable length UTF-8 encoded string.
     *
     * @see <a href="https://wiki.vg/Protocol#Data_types">https://wiki.vg/Protocol#Data_types</a>
     */
    fun ByteBuf.readString(): String {
        val length = readVarInt()
        val bytes = ByteArray(length)
        readBytes(bytes)
        return bytes.toString(Charsets.UTF_8)
    }

    /**
     * Writes a variable length UTF-8 encoded string.
     *
     * @see <a href="https://wiki.vg/Protocol#Data_types">https://wiki.vg/Protocol#Data_types</a>
     */
    fun ByteBuf.writeString(value: String): ByteBuf {
        val bytes = value.toByteArray(Charsets.UTF_8)
        writeVarInt(bytes.size)
        writeBytes(bytes)
        return this
    }

    /**
     * Reads a UUID encoded as two 64-bit longs.
     *
     * @see <a href="https://wiki.vg/Protocol#Data_types">https://wiki.vg/Protocol#Data_types</a>
     */
    fun ByteBuf.readUUID(): UUID {
        return UUID(readLong(), readLong())
    }

    /**
     * Writes a UUID as two 64-bit longs.
     *
     * @see <a href="https://wiki.vg/Protocol#Data_types">https://wiki.vg/Protocol#Data_types</a>
     */
    fun ByteBuf.writeUUID(value: UUID): ByteBuf {
        writeLong(value.mostSignificantBits)
        writeLong(value.leastSignificantBits)
        return this
    }

    /**
     * Writes a compound NBT tag with an empty name.
     */
    fun ByteBuf.writeNBT(value: NBT): ByteBuf {
        DataOutputStream(ByteBufOutputStream(this)).use {
            value.write(it, "")
            it.flush()
        }

        return this
    }

    /**
     * Writes a [VoxelLocation] encoded as long.
     */
    fun ByteBuf.writeVoxelLocation(value: VoxelLocation): ByteBuf {
        writeLong(
            value.x.toLong() and 0x3FFFFFF shl 38 or
                    (value.z.toLong() and 0x3FFFFFF shl 12) or
                    (value.y.toLong() and 0xFFF)
        )

        return this
    }
}
