package space.uranos.net.packet.play

import io.netty.buffer.ByteBuf
import space.uranos.chat.TextComponent
import space.uranos.net.MinecraftProtocolDataTypes.writeString
import space.uranos.net.MinecraftProtocolDataTypes.writeUUID
import space.uranos.net.MinecraftProtocolDataTypes.writeVarInt
import space.uranos.net.packet.OutgoingPacketCodec
import space.uranos.player.GameMode

object PlayerInfoPacketCodec : OutgoingPacketCodec<PlayerInfoPacket>(0x32, PlayerInfoPacket::class) {
    override fun PlayerInfoPacket.encode(dst: ByteBuf) {
        val encoder = when (action) {
            is PlayerInfoPacket.Action.AddPlayer -> ActionEncoder.AddPlayer
            is PlayerInfoPacket.Action.UpdateGameMode -> ActionEncoder.UpdateGameMode
            is PlayerInfoPacket.Action.UpdateLatency -> ActionEncoder.UpdateLatency
            is PlayerInfoPacket.Action.UpdateDisplayName -> ActionEncoder.UpdateDisplayName
            is PlayerInfoPacket.Action.RemovePlayer -> ActionEncoder.RemovePlayer
        }

        dst.writeVarInt(encoder.id)
        dst.writeVarInt(action.entries.size)

        action.entries.forEach { (uuid, value) ->
            dst.writeUUID(uuid)
            encoder.encodeUnsafeWithCast(dst, value)
        }
    }

    sealed class ActionEncoder<T> {
        abstract val id: Int
        abstract fun encode(dst: ByteBuf, value: T)
        fun encodeUnsafeWithCast(dst: ByteBuf, value: Any?) {
            @Suppress("UNCHECKED_CAST")
            encode(dst, value as T)
        }

        object AddPlayer: ActionEncoder<PlayerInfoPacket.Action.AddPlayer.Data>() {
            override val id: Int = 0
            override fun encode(dst: ByteBuf, value: PlayerInfoPacket.Action.AddPlayer.Data) {
                dst.writeString(value.name)
                dst.writeVarInt(value.properties.size)

                for ((name, property) in value.properties) {
                    dst.writeString(name)
                    dst.writeString(property.value)

                    val hasSignature = property.signature != null
                    dst.writeBoolean(hasSignature)
                    if (hasSignature) dst.writeString(property.signature!!)
                }

                dst.writeVarInt(value.gameMode.numericID)
                dst.writeVarInt(value.latency)

                if (value.displayName != null) {
                    dst.writeBoolean(true)
                    dst.writeString(value.displayName!!.toJson())
                } else dst.writeBoolean(false)
            }
        }

        object UpdateGameMode: ActionEncoder<GameMode>() {
            override val id: Int = 1
            override fun encode(dst: ByteBuf, value: GameMode) {
                dst.writeVarInt(value.numericID)
            }
        }

        object UpdateLatency: ActionEncoder<Int>() {
            override val id: Int = 2
            override fun encode(dst: ByteBuf, value: Int) {
                dst.writeVarInt(value)
            }
        }

        object UpdateDisplayName: ActionEncoder<TextComponent?>() {
            override val id: Int = 3
            override fun encode(dst: ByteBuf, value: TextComponent?) {
                if (value == null) dst.writeBoolean(false)
                else {
                    dst.writeBoolean(true)
                    dst.writeString(value.toJson())
                }
            }
        }

        object RemovePlayer: ActionEncoder<Unit>() {
            override val id: Int = 4
            override fun encode(dst: ByteBuf, value: Unit) {}
        }
    }
}
