package space.uranos.net.packet.play

import space.uranos.chat.TextComponent
import space.uranos.net.packet.Mergeable
import space.uranos.net.packet.OutgoingPacket
import space.uranos.player.GameMode
import java.util.UUID

/**
 * Informs the client about other players.
 */
data class PlayerInfoPacket(val action: Action<*>) : OutgoingPacket(), Mergeable {
    sealed class Action<T> {
        abstract val entries: Map<UUID, T>

        data class AddPlayer(override val entries: Map<UUID, Data>) : Action<AddPlayer.Data>() {
            data class Data(
                val name: String,
                /**
                 * If this does not match the real game mode of a player, weird things can happen.
                 *
                 * Use with caution.
                 */
                val gameMode: GameMode,
                /**
                 * Latency displayed in the player list in milliseconds.
                 *
                 * See [LATENCY] for constants.
                 */
                val latency: Int,
                /**
                 * Only visible in the player list.
                 */
                val displayName: TextComponent?,
                /**
                 * Primarily used for the player skin (key: textures).
                 */
                val properties: Map<String, Property>
            ) {
                data class Property(
                    val value: String,
                    val signature: String? = null
                )
            }
        }

        /**
         * See [AddPlayer.Data.gameMode].
         */
        data class UpdateGameMode(override val entries: Map<UUID, GameMode>) : Action<GameMode>()

        /**
         * See [AddPlayer.Data.latency].
         */
        data class UpdateLatency(override val entries: Map<UUID, Int>) : Action<Int>()

        /**
         * See [AddPlayer.Data.displayName].
         */
        data class UpdateDisplayName(override val entries: Map<UUID, TextComponent?>) : Action<TextComponent?>()

        data class RemovePlayer(override val entries: Map<UUID, Unit>) : Action<Unit>() {
            constructor(entries: List<UUID>) : this(entries.map { it to Unit }.toMap())
        }
    }

    override fun mergeWith(otherPacket: OutgoingPacket): OutgoingPacket? {
        if (otherPacket !is PlayerInfoPacket) return null

        val action = when (otherPacket.action) {
            is Action.AddPlayer -> (action as? Action.AddPlayer)?.let { Action.AddPlayer(otherPacket.action.entries + it.entries) }
            is Action.UpdateGameMode -> (action as? Action.UpdateGameMode)?.let { Action.UpdateGameMode(otherPacket.action.entries + it.entries) }
            is Action.UpdateLatency -> (action as? Action.UpdateLatency)?.let { Action.UpdateLatency(otherPacket.action.entries + it.entries) }
            is Action.UpdateDisplayName -> (action as? Action.UpdateDisplayName)?.let {
                Action.UpdateDisplayName(
                    otherPacket.action.entries + it.entries
                )
            }
            is Action.RemovePlayer -> (action as? Action.RemovePlayer)?.let { Action.RemovePlayer(otherPacket.action.entries + it.entries) }
        }

        return action?.let { PlayerInfoPacket(it) }
    }

    /**
     * Latency constants for use in the [AddPlayer][Action.AddPlayer] and [UpdateLatency][Action.UpdateLatency] actions.
     */
    object LATENCY {
        const val NO_CONNECTION = -1
        const val FIVE_BARS = 0
        const val FOUR_BARS = 150
        const val THREE_BARS = 300
        const val TWO_BARS = 600
        const val ONE_BAR = 1000
    }
}
