package space.uranos

import kotlinx.coroutines.CancellableContinuation
import kotlinx.coroutines.runBlocking
import kotlinx.coroutines.suspendCancellableCoroutine
import space.uranos.logging.Logger
import space.uranos.server.Server
import java.util.Collections
import java.util.concurrent.*
import kotlin.coroutines.resume

/**
 * Basically ExecutorService but for coroutines and with ticks.
 */
class UranosScheduler : Scheduler {
    private val executor: ScheduledExecutorService =
        Executors.newSingleThreadScheduledExecutor { r -> Thread(r, "Scheduler") }

    private val tasks = ConcurrentHashMap.newKeySet<Task<out Any>>()
    private val shutdownTasks = Collections.synchronizedSet(LinkedHashSet<suspend () -> Unit>())

    val logger = Logger("Scheduler", false)

    inner class Task<R : Any>(
        val fn: suspend () -> R,
        val interval: Long?,
        var ticksUntilExecution: Long
    ) : Scheduler.Task {
        @Volatile
        var cancelled: Boolean = false

        val creationStackTrace = Thread.currentThread().stackTrace

        override fun cancel() {
            tasks.remove(this)
            cancelled = true
        }
    }

    private var future: ScheduledFuture<*>? = null

    fun start() {
        if (future != null) throw IllegalStateException("Already started")
        val interval = 1000L / Server.TICKS_PER_SECOND

        future = executor.scheduleAtFixedRate({
            runBlocking {
                val startTime = System.currentTimeMillis()

                for (task in tasks) {
                    if (!task.cancelled) {
                        if (task.ticksUntilExecution <= 0) {
                            try {
                                task.fn()

                                if (task.interval != null) task.ticksUntilExecution = task.interval
                                else tasks.remove(task)
                            } catch (e: Exception) {
                                tasks.remove(task)
                                e.printStackTrace()
                            }
                        } else {
                            task.ticksUntilExecution -= 1
                        }
                    }
                }

                val duration = System.currentTimeMillis() - startTime
                if (duration > interval) logger.warn("Last tick took ${duration}ms, but should only take ${interval}ms")
            }
        }, 0, interval, TimeUnit.MILLISECONDS)
    }

    private fun stop() {
        if (future == null) throw IllegalArgumentException("Not started")
        future!!.cancel(false)
        future = null
    }

    suspend fun shutdown() {
        stop()
        shutdownTasks.forEach { it.invoke() }
    }

    override fun executeAfter(delay: Long, block: suspend () -> Unit): Scheduler.Task {
        val task = Task(block, null, delay)
        tasks.add(task)
        return task
    }

    override fun executeRepeating(interval: Long, delay: Long, block: suspend () -> Unit): Scheduler.Task {
        val task = Task(block, interval, delay)
        tasks.add(task)
        return task
    }

    override fun registerShutdownTask(block: suspend () -> Unit): Scheduler.Task {
        shutdownTasks.add(block)

        return object : Scheduler.Task {
            override fun cancel() {
                shutdownTasks.remove(block)
            }
        }
    }

    override suspend fun <R> runAfter(delay: Long, block: suspend () -> R): R {
        lateinit var continuation: CancellableContinuation<R>

        val fn = suspend { continuation.resume(block()) }
        val task = Task(fn, null, delay)

        return suspendCancellableCoroutine {
            it.invokeOnCancellation {
                task.cancel()
            }

            continuation = it
            tasks.add(task)
        }
    }
}
