package space.uranos

import com.sksamuel.hoplite.ConfigFilePropertySource
import com.sksamuel.hoplite.ConfigLoader
import com.sksamuel.hoplite.ConfigSource
import kotlinx.coroutines.runBlocking
import space.uranos.config.UranosConfig
import space.uranos.entity.*
import space.uranos.entity.impl.*
import space.uranos.event.UranosEventBus
import space.uranos.event.UranosEventHandlerPositionManager
import space.uranos.logging.Logger
import space.uranos.logging.UranosLoggingOutputProvider
import space.uranos.net.UranosSocketServer
import space.uranos.net.packet.play.PlayerInfoPacket
import space.uranos.player.UranosPlayer
import space.uranos.plugin.UranosPluginManager
import space.uranos.server.Server
import space.uranos.util.EncryptionUtils
import space.uranos.util.msToTicks
import space.uranos.util.runInServerThread
import space.uranos.world.UranosWorldRegistry
import java.io.File
import java.security.KeyPair
import java.util.concurrent.atomic.AtomicInteger
import kotlin.system.exitProcess

// TODO: Consider using DI because this improves testability

class UranosServer internal constructor() : Server() {
    val logger = Logger("Server")
    private val socketServer = UranosSocketServer(this)

    val keyPair: KeyPair =
        try {
            EncryptionUtils.generateKeyPair()
        } catch (e: Exception) {
            failInitialization(e)
        }

    val x509EncodedPublicKey: ByteArray = EncryptionUtils.generateX509Key(keyPair.public).encoded

    override val sessions by socketServer::sessions
    override val players get() = sessions.mapNotNull { it.player as UranosPlayer? }

    override val pluginManager = UranosPluginManager(this)
    override val serverDirectory: File =
        File(UranosServer::class.java.protectionDomain.codeSource.location.toURI()).let {
            it.mkdirs()
            if (VERSION == "development") it.resolve("../../../../../serverData").normalize()
            else it
        }

    override val loggingOutputProvider = UranosLoggingOutputProvider
    override val scheduler = UranosScheduler()

    val config = ConfigLoader.Builder()
        .addPropertySource(
            ConfigFilePropertySource(
                ConfigSource.FileSource(serverDirectory.resolve("./config.yml").normalize()),
                optional = true
            )
        )
        .addPropertySource(
            ConfigFilePropertySource(
                ConfigSource.FileSource(File(javaClass.classLoader.getResource("default-config.yml")!!.toURI()))
            )
        )
        .build().loadConfigOrThrow<UranosConfig>()

    override val minimumLogLevel = config.logging.minLevel
    override val developmentMode: Boolean = config.developmentMode

    override val eventBus = UranosEventBus(config.logging.events)
    override val eventHandlerPositions = UranosEventHandlerPositionManager()

    override val worldRegistry = UranosWorldRegistry()

    private val internalEntities = HashSet<UranosEntity>()
    override val entities: Set<Entity> = internalEntities

    override fun <T : Entity> create(type: EntityType<T>): T {
        val entity: UranosEntity = when (type) {
            CowEntity -> UranosCowEntity(this)
            BatEntity -> UranosBatEntity(this)
            CreeperEntity -> UranosCreeperEntity(this)
            MinecartEntity -> UranosMinecartEntity(this)
            else -> throw IllegalArgumentException("Entities of this type cannot be created with this function")
        }

        internalEntities.add(entity)

        @Suppress("UNCHECKED_CAST")
        return entity as T
    }

    fun createPlayerEntity(player: UranosPlayer) =
        UranosPlayerEntity(this, player).also { internalEntities.add(it) }

    override fun shutdown() {
        runBlocking {
            scheduler.shutdown()
        }
    }

    private val nextEntityID = AtomicInteger(1)
    fun claimEntityID() = nextEntityID.getAndIncrement()

    private fun failInitialization(t: Throwable): Nothing {
        logger.error("Server initialization failed:", t)
        exitProcess(1)
    }

    suspend fun start() {
        logger info "Starting UranosServer ($VERSION_WITH_V)"
        logger trace "Configuration: $config"

        pluginManager.loadAll()
        pluginManager.enableAll()

        socketServer.bind()
        logger info "Listening on ${config.host}:${config.port}"

        scheduler.start()
        registerListeners()
        startTicking()
        startPingSync()
    }

    private fun startTicking() {
        var index = 0
        scheduler.executeRepeating(1, 0) {
            runInServerThread {
                players.forEach { it.container.tick() }
                internalEntities.forEach { it.tick() }
                sessions.forEach { it.tick(index) }
            }

            index++
        }
    }

    private fun startPingSync() {
        // Not in UranosSession.tick because this simplifies reusing the packet and the ping is not a critical information
        scheduler.executeRepeating(msToTicks(config.pingUpdateInterval.toMillis()), 0) {
            val packet = PlayerInfoPacket(
                PlayerInfoPacket.Action.UpdateLatency(players.map { it.uuid to it.session.ping }.toMap())
            )

            players.forEach {
                it.session.send(packet)
            }
        }
    }

    private fun registerListeners() {
        // Nothing
    }

    companion object {
        val VERSION = UranosServer::class.java.`package`.implementationVersion ?: "development"
        val VERSION_WITH_V = if (VERSION == "development") VERSION else "v$VERSION"

        @JvmStatic
        fun main(args: Array<String>) = runBlocking {
            UranosServer().also { setServerInstance(it) }.start()
        }

        fun setServerInstance(instance: Server) {
            val clazz = Class.forName("space.uranos.UranosKt")
            val field = clazz.getDeclaredField("Uranos")
            field.isAccessible = true
            field.set(null, instance)
            field.isAccessible = false
        }
    }
}
