package space.uranos.entity

import kotlinx.coroutines.launch
import kotlinx.coroutines.sync.Mutex
import kotlinx.coroutines.sync.withLock
import space.uranos.*
import space.uranos.entity.event.ViewingChangedEvent
import space.uranos.entity.impl.UranosPlayerEntity
import space.uranos.net.packet.OutgoingPacket
import space.uranos.net.packet.play.*
import space.uranos.player.Player
import space.uranos.util.*
import space.uranos.util.collections.WatchableSet
import space.uranos.util.numbers.mapToUByte
import space.uranos.util.numbers.validatePitch
import space.uranos.util.numbers.validateYaw
import space.uranos.world.Chunk
import space.uranos.world.VoxelLocation
import space.uranos.world.World
import space.uranos.world.getInternalsIfRegistered
import java.util.Collections
import java.util.UUID
import java.util.WeakHashMap

sealed class UranosEntity(server: UranosServer) : Entity {
    override fun belongsToChunk(key: Chunk.Key): Boolean = key == chunkKey

    override val numericID: Int = server.claimEntityID()
    override val uuid: UUID = UUID.randomUUID()

    override var visibleToNewPlayers: Boolean = true

    private val worldMutex = Mutex()
    final override var world: World? = null; private set

    override suspend fun setWorld(world: World?) {
        if (world == null && this is PlayerEntity)
            throw IllegalArgumentException("You cannot set the world of a PlayerEntity to null")

        if (world == this.world) return

        worldMutex.withLock {
            this.world?.getInternalsIfRegistered()?.removeEntity(this)
            this.world = world
            world?.getInternalsIfRegistered()?.addEntity(this)
        }
    }

    protected val addedViewers = mutableSetOf<Player>()
    protected val removedViewers = mutableSetOf<Player>()

    override val viewers: MutableSet<Player> = object : WatchableSet<Player>(Collections.newSetFromMap(WeakHashMap())) {
        override fun beforeAdd(element: Player) {
            if ((this@UranosEntity as? UranosPlayerEntity)?.let { it.player == element } == true)
                throw IllegalArgumentException("A player cannot be a viewer of it's own entity")
        }

        override fun onAdd(element: Player) {
            removedViewers.remove(element)
            addedViewers.add(element)
            Uranos.scope.launch { Uranos.eventBus.emit(ViewingChangedEvent(this@UranosEntity, element, true)) }
        }

        override fun onRemove(element: Player) {
            addedViewers.remove(element)
            removedViewers.add(element)
            Uranos.scope.launch { Uranos.eventBus.emit(ViewingChangedEvent(this@UranosEntity, element, false)) }
        }
    }

    abstract suspend fun tick()
    abstract val chunkKey: Chunk.Key

    protected val container = TickSynchronizationContainer()

    protected fun sendSpawnAndDestroyPackets() {
        if (addedViewers.isNotEmpty()) createSpawnPacket().let { packet -> addedViewers.forEach { it.session.send(packet) } }
        if (removedViewers.isNotEmpty()) DestroyEntitiesPacket(arrayOf(numericID)).let { packet -> removedViewers.forEach { it.session.send(packet) } }
    }

    protected fun finishTick() {
        addedViewers.clear()
        removedViewers.clear()
    }
}

sealed class UranosLivingEntity(server: UranosServer) : UranosEntity(server), LivingEntity {
    override var velocity: Vector = Vector.ZERO
    override var position: Position = Position.ZERO
    override val chunkKey: Chunk.Key by memoized({ position }) { Chunk.Key.from(position) }
}

abstract class UranosNotHasMovableHeadLivingEntity(server: UranosServer) : UranosLivingEntity(server) {
    override var velocity: Vector = Vector.ZERO
    override var position: Position = Position.ZERO

    override val chunkKey: Chunk.Key by memoized({ position }) { Chunk.Key.from(position) }

    private var lastSentPosition: Position = Position.ZERO
    private var lastSentYaw: Float = 0f
    private var lastSentPitch: Float = 0f

    final override suspend fun tick() {
        container.tick()

        val viewersWithoutAdded = viewers.subtract(addedViewers)
        if (viewersWithoutAdded.isNotEmpty()) {
            val packet = createMovementPacket()
            if (packet != null) {
                viewersWithoutAdded.forEach { it.session.send(packet) }
            }
        }

        if (this is PitchRotatable) lastSentPitch = pitch
        if (this is YawRotatable) lastSentYaw = yaw

        sendSpawnAndDestroyPackets()
        finishTick()
    }

    private fun createMovementPacket(): OutgoingPacket? = createNotHasMovableHeadMovementPacket(position, lastSentPosition, lastSentYaw, lastSentPitch)
}

abstract class UranosHasMovableHeadLivingEntity(server: UranosServer) : UranosLivingEntity(server), HasMovableHead {
    override var headYaw: Float = 0f
        set(value) {
            validateYaw(value, "headYaw"); field = value
        }

    override var headPitch: Float = 0f
        set(value) {
            validatePitch(value, "headPitch"); field = value
        }

    final override suspend fun tick() {
        container.tick()

        val viewersWithoutAdded = viewers.subtract(addedViewers)
        if (viewersWithoutAdded.isNotEmpty()) {
            val packets = createMovementPackets()
            if (packets.isNotEmpty()) {
                viewersWithoutAdded.forEach { it.session.send(packets) }
            }
        }

        oldPosition = position
        oldHeadPitch = headPitch
        oldHeadYaw = headYaw

        sendSpawnAndDestroyPackets()
        finishTick()
    }

    private var oldPosition: Position = Position.ZERO
    private var oldHeadYaw: Float = 0f
    private var oldHeadPitch: Float = 0f

    private fun createMovementPackets(): ArrayList<OutgoingPacket> {
        val packets = ArrayList<OutgoingPacket>(2)

        val delta = position.asVector() - oldPosition.asVector()
        val absoluteDelta = abs(delta)
        val onGround = true

        val oldHeadPitchUByte = oldHeadPitch.mapToUByte(90f)
        val newHeadPitchUByte = headPitch.mapToUByte(90f)
        val oldHeadYawUByte = oldHeadYaw.mapToUByte(360f)
        val newHeadYawUByte = headYaw.mapToUByte(360f)

        if (absoluteDelta.x + absoluteDelta.y + absoluteDelta.z == 0.0) {
            if (oldHeadPitchUByte != newHeadPitchUByte) packets += EntityOrientationPacket(
                numericID,
                0u,
                newHeadPitchUByte,
                onGround
            )
        } else if (absoluteDelta.x > 8 || absoluteDelta.y > 8 || absoluteDelta.z > 8) {
            packets += EntityTeleportPacket(numericID, position.x, position.y, position.z, 0u, newHeadPitchUByte, onGround)
        } else {
            packets += if (oldHeadPitchUByte != newHeadPitchUByte) EntityRelativeMoveWithOrientationPacket(
                numericID,
                EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.x),
                EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.y),
                EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.z),
                0u,
                newHeadPitchUByte,
                onGround
            ) else EntityRelativeMovePacket(
                numericID,
                EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.x),
                EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.y),
                EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.z),
                onGround
            )
        }

        if (oldHeadYawUByte != newHeadYawUByte) packets += EntityHeadYawPacket(numericID, newHeadYawUByte)

        return packets
    }
}

abstract class UranosObjectEntity(server: UranosServer) : UranosEntity(server), ObjectEntity {
    override var velocity: Vector = Vector.ZERO
    override var position: Position = Position.ZERO

    override val chunkKey: Chunk.Key by memoized({ position }) { Chunk.Key.from(position) }

    private var lastSentPosition: Position = Position.ZERO
    private var lastSentYaw: Float = 0f
    private var lastSentPitch: Float = 0f

    final override suspend fun tick() {
        container.tick()

        val viewersWithoutAdded = viewers.subtract(addedViewers)
        if (viewersWithoutAdded.isNotEmpty()) {
            val packet = createMovementPacket()
            if (packet != null) {
                viewersWithoutAdded.forEach { it.session.send(packet) }
            }
        }

        if (this is PitchRotatable) lastSentPitch = pitch
        if (this is YawRotatable) lastSentYaw = yaw

        sendSpawnAndDestroyPackets()
        finishTick()
    }

    private fun createMovementPacket(): OutgoingPacket? = createNotHasMovableHeadMovementPacket(position, lastSentPosition, lastSentYaw, lastSentPitch)
}

class UranosPaintingEntity(
    server: UranosServer,
    override val topLeftLocation: VoxelLocation,
    override val facing: CardinalDirection,
    override val motive: PaintingMotive
) : UranosEntity(server), PaintingEntity {
    override val chunkKey: Chunk.Key get() = Chunk.Key.from(topLeftLocation)

    private var lastSentTopLeftLocation: VoxelLocation = topLeftLocation

    override suspend fun tick() {
        container.tick()

        if (lastSentTopLeftLocation != topLeftLocation) {
            val viewersWithoutAdded = viewers.subtract(addedViewers)
            if (viewersWithoutAdded.isNotEmpty()) {
                val centerLocation = calculateCenterLocation()
                val packet = EntityTeleportPacket(
                    numericID,
                    centerLocation.x.toDouble(),
                    centerLocation.y.toDouble(),
                    centerLocation.z.toDouble(),
                    0u,
                    0u,
                    true
                )

                viewersWithoutAdded.forEach { it.session.send(packet) }
            }
        }

        lastSentTopLeftLocation = topLeftLocation

        sendSpawnAndDestroyPackets()
        finishTick()
    }
}

fun UranosEntity.createNotHasMovableHeadMovementPacket(
    position: Position,
    lastSentPosition: Position,
    lastSentYaw: Float,
    lastSentPitch: Float
): OutgoingPacket? {
    val delta = position.asVector() - lastSentPosition.asVector()
    val absoluteDelta = abs(delta)
    val onGround = true

    val oldPitchUByte = lastSentPitch.mapToUByte(90f)
    val newPitchUByte = if (this is PitchRotatable) pitch.mapToUByte(90f) else 0u
    val oldYawUByte = lastSentYaw.mapToUByte(360f)
    val newYawUByte = if (this is YawRotatable) yaw.mapToUByte(360f) else 0u

    return if (absoluteDelta.x + absoluteDelta.y + absoluteDelta.z == 0.0) {
        if (oldPitchUByte != newPitchUByte || oldYawUByte != newYawUByte) EntityOrientationPacket(
            numericID,
            newYawUByte,
            newPitchUByte,
            onGround
        ) else null
    } else if (absoluteDelta.x > 8 || absoluteDelta.y > 8 || absoluteDelta.z > 8) {
        EntityTeleportPacket(numericID, position.x, position.y, position.z, newYawUByte, newPitchUByte, onGround)
    } else {
        if (oldPitchUByte != newPitchUByte || oldYawUByte != newYawUByte) EntityRelativeMoveWithOrientationPacket(
            numericID,
            EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.x),
            EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.y),
            EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.z),
            newYawUByte,
            newPitchUByte,
            onGround
        ) else EntityRelativeMovePacket(
            numericID,
            EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.x),
            EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.y),
            EntityRelativeMoveWithOrientationPacket.convertToDeltaShort(delta.z),
            onGround
        )
    }
}
