package space.uranos.net

import io.netty.buffer.Unpooled
import space.uranos.Difficulty
import space.uranos.DifficultySettings
import space.uranos.UranosServer
import space.uranos.chat.TextComponent
import space.uranos.event.ifCancelled
import space.uranos.net.MinecraftProtocolDataTypes.writeString
import space.uranos.net.event.PlayerInitializationEvent
import space.uranos.net.event.SessionAfterLoginEvent
import space.uranos.net.packet.login.*
import space.uranos.net.packet.play.*
import space.uranos.player.UranosPlayer
import space.uranos.player.event.PlayerReadyEvent
import space.uranos.tag.TagRegistry
import space.uranos.util.AuthenticationHelper
import space.uranos.util.EncryptionUtils
import space.uranos.world.Chunk
import java.security.MessageDigest
import java.util.UUID
import javax.crypto.Cipher
import javax.crypto.spec.SecretKeySpec

class LoginAndJoinProcedure(val session: UranosSession) {
    private val tagsPacket by lazy { TagsPacket(TagRegistry.tags) }

    suspend fun start(packet: LoginStartPacket) {
        session.state.getOrFail<Session.State.WaitingForLoginStart>()

        if (session.server.config.authenticateAndEncrypt) {
            val verifyToken = EncryptionUtils.generateVerifyToken()
            session.state = Session.State.WaitingForEncryptionVerification(packet.username, verifyToken)
            session.sendNow(EncryptionRequestPacket(session.server.x509EncodedPublicKey, verifyToken))
        } else {
            val uuid = UUID.nameUUIDFromBytes("OfflinePlayer:${packet.username}".toByteArray())
            session.sendNow(LoginSuccessPacket(uuid, packet.username))
            session.state = Session.State.LoginSucceeded(packet.username, uuid)
            afterLogin()
        }
    }

    suspend fun onEncryptionResponse(packet: EncryptionResponsePacket) {
        val state: Session.State.WaitingForEncryptionVerification = session.state.getOrFail()

        val cipher: Cipher = Cipher.getInstance("RSA")
        cipher.init(Cipher.DECRYPT_MODE, session.server.keyPair.private)

        // Decrypt verify token
        val verifyToken: ByteArray = cipher.doFinal(packet.verifyToken)

        if (!verifyToken.contentEquals(state.verifyToken)) {
            session.disconnect(TextComponent of "Encryption failed.")
            return
        }

        // Decrypt shared secret
        val sharedSecretKey = SecretKeySpec(cipher.doFinal(packet.sharedSecret), "AES")

        session.state = Session.State.Encrypted(state.username)
        session.enableEncryptionCodec(sharedSecretKey)

        val hash = MessageDigest.getInstance("SHA-1")
        hash.update(sharedSecretKey.encoded)
        hash.update(session.server.x509EncodedPublicKey)
        val hashString = EncryptionUtils.toMinecraftStyleSha1String(hash.digest())

        val result = AuthenticationHelper.authenticate(hashString, state.username)

        session.sendNow(SetCompressionPacket(session.server.config.packetCompressionThreshold))
        session.enableCompressionCodec()

        session.sendNow(LoginSuccessPacket(result.uuid, result.username))
        session.state = Session.State.LoginSucceeded(result.username, result.uuid)
        afterLogin()
    }

    private suspend fun afterLogin() {
        val state: Session.State.LoginSucceeded = session.state.getOrFail()
        session.packetsAdapter.launchPacketsChannelConsumer()

        val event = session.server.eventBus.emit(SessionAfterLoginEvent(session))
        val initialWorldAndLocation = event.initialWorldAndLocation

        when {
            event.cancelled -> session.disconnect()
            initialWorldAndLocation == null -> {
                session.logger warn "Could not join because no spawn location was set"
                session.disconnect(internalReason = "No spawn location set")
            }
            else -> {
                session.sendNow(
                    JoinGamePacket(
                        0,
                        event.gameMode,
                        event.hardcoreHearts,
                        initialWorldAndLocation.first,
                        event.maxViewDistance,
                        event.reducedDebugInfo,
                        event.respawnScreenEnabled
                    )
                )

                session.sendPluginMessage(
                    "minecraft:brand",
                    Unpooled.buffer().writeString("Uranos ${UranosServer.VERSION_WITH_V}")
                )

                // As this is only visual, there is no way of changing it aside from intercepting the packet.
                session.sendNow(ServerDifficultyPacket(DifficultySettings(Difficulty.NORMAL, false)))

                session.sendNow(
                    PlayerAbilitiesPacket(
                        event.invulnerable,
                        event.flying,
                        event.canFly,
                        // TODO: Find out how this relates to the entity property named `generic.flying_speed`
                        event.flyingSpeed,
                        event.fieldOfView
                    )
                )

                session.state = Session.State.WaitingForClientSettings(
                    state.username,
                    state.uuid,
                    event.flying,
                    event.canFly,
                    event.flyingSpeed,
                    event.fieldOfView,
                    event.gameMode,
                    initialWorldAndLocation.first,
                    initialWorldAndLocation.second,
                    event.headYaw,
                    event.headPitch,
                    event.invulnerable,
                    event.reducedDebugInfo,
                    event.selectedHotbarSlot
                )
            }
        }
    }

    suspend fun onClientSettingsReceived(packet: ClientSettingsPacket) {
        val state: Session.State.WaitingForClientSettings = session.state.getOrFail()
        val settings = packet.asPlayerSettings()

        val event = session.server.eventBus.emit(PlayerInitializationEvent(session, settings))

        event.ifCancelled {
            session.disconnect(internalReason = "PlayerInitializationEvent was cancelled")
            return
        }

        val player = UranosPlayer(
            session,
            state.username,
            state.uuid,
            state.gameMode,
            settings,
            state.reducedDebugInfo,
            state.fieldOfView,
            state.canFly,
            state.flying,
            state.flyingSpeed,
            state.invulnerable,
            event.compassTarget,
            state.selectedHotbarSlot
        )

        session.state = Session.State.Joining(player)

        session.sendNow(SelectedHotbarSlotPacket(state.selectedHotbarSlot))

        session.sendNow(DeclareRecipesPacket(session.server.recipeRegistry.items.values))
        session.sendNow(tagsPacket)

//        session.send(DeclareCommandsPacket(session.server.commandRegistry.items.values))
        // UnlockRecipes

        session.sendNow(OutgoingPlayerPositionPacket(state.position.x, state.position.y, state.position.z, state.headYaw, state.headPitch))

        session.sendNow(PlayerInfoPacket(PlayerInfoPacket.Action.AddPlayer((session.server.players + player).map {
            it.uuid to PlayerInfoPacket.Action.AddPlayer.Data(
                it.name,
                it.gameMode,
                it.session.ping,
                it.playerListName,
                emptyMap() // TODO: Load skin
            )
        }.toMap())))

        session.sendNow(
            PlayerInfoPacket(
                PlayerInfoPacket.Action.UpdateLatency(
                    session.server.players.map { it.uuid to it.session.ping }.toMap()
                )
            )
        )

        session.sendNow(ViewPositionPacket(Chunk.Key.from(player.entity.position.toVoxelLocation())))

        session.sendKeepAlivePacket()

        player.spawnInitially(state.world)
        session.state = Session.State.Playing(player)

        session.server.eventBus.emit(PlayerReadyEvent(session.player!!))

        // WorldBorder
        session.send(CompassTargetPacket(player.compassTarget))
    }
}
