package space.uranos.net

import io.netty.buffer.ByteBuf
import kotlinx.coroutines.*
import space.uranos.UranosServer
import space.uranos.chat.ChatColor
import space.uranos.chat.ChatComponent
import space.uranos.chat.TextComponent
import space.uranos.event.*
import space.uranos.logging.Logger
import space.uranos.net.packet.OutgoingPacket
import space.uranos.net.packet.Protocol
import space.uranos.net.packet.handshaking.HandshakingProtocol
import space.uranos.net.packet.login.LoginProtocol
import space.uranos.net.packet.play.OutgoingKeepAlivePacket
import space.uranos.net.packet.play.OutgoingPluginMessagePacket
import space.uranos.net.packet.play.PlayProtocol
import space.uranos.net.packet.play.PlayerInfoPacket
import space.uranos.net.packet.status.StatusProtocol
import space.uranos.server.event.SessionInitializedEvent
import space.uranos.util.msToTicks
import java.net.InetAddress
import java.net.InetSocketAddress
import javax.crypto.SecretKey

class UranosSession(val channel: io.netty.channel.Channel, val server: UranosServer) : Session() {
    override val address: InetAddress = (channel.remoteAddress() as InetSocketAddress).address

    private val identifier = "UranosSession(${address.hostAddress})"
    val logger = Logger(identifier)

    override var brand: String? = null

    override var ping: Int = -1
        set(value) {
            val player = earlyPlayer

            if (player != null && field == -1) {
                val packet = PlayerInfoPacket(
                    PlayerInfoPacket.Action.UpdateLatency(mapOf(player.uuid to value))
                )

                scope.launch { server.players.forEach { it.session.send(packet) } }
            }

            field = value
        }

    val earlyPlayer get() = (state as? State.WithPlayer)?.player

    override var state: State = State.WaitingForHandshake

    override val currentProtocol: Protocol?
        get() = when (state) {
            is State.WaitingForHandshake -> HandshakingProtocol

            is State.WaitingForStatusRequest -> StatusProtocol

            is State.WaitingForLoginStart,
            is State.WaitingForEncryptionVerification,
            is State.Encrypted -> LoginProtocol

            is State.LoginSucceeded,
            is State.WaitingForClientSettings,
            is State.Joining,
            is State.Playing -> PlayProtocol

            is State.Disconnected -> null
        }

    val packetsAdapter = PacketsAdapter(this)

    override suspend fun sendNow(packet: OutgoingPacket) = packetsAdapter.send(packet)
    override fun send(packet: OutgoingPacket) = packetsAdapter.sendNextTick(packet)

    override suspend fun sendPluginMessage(channel: String, data: ByteBuf) {
        if (this.currentProtocol != PlayProtocol) throw IllegalStateException("The session is not using the PLAY protocol")
        sendNow(OutgoingPluginMessagePacket(channel, data))
        data.release()
    }

    override suspend fun disconnect(reason: TextComponent, internalReason: String?) {
        packetsAdapter.stopProcessingIncomingPackets()

        if (currentProtocol == LoginProtocol || currentProtocol == PlayProtocol) {
            logger info "Disconnected" + (internalReason?.let { ". Internal reason: $it" } ?: "")

            val finalReason = if (server.config.developmentMode && internalReason != null) {
                val additional = listOf<ChatComponent>(
                    TextComponent(
                        "\n\nLogged reason (only shown in development mode): \n",
                        color = ChatColor.GRAY
                    ),
                    TextComponent of internalReason
                )

                reason.copy(extra = reason.extra + additional)
            } else reason

            when (currentProtocol) {
                LoginProtocol -> sendNow(space.uranos.net.packet.login.DisconnectPacket(finalReason))
                PlayProtocol -> sendNow(space.uranos.net.packet.play.DisconnectPacket(finalReason))
            }
        } else logger trace "Disconnected"

        onDisconnect(true)
        channel.close()
    }

    val joinProcedure = LoginAndJoinProcedure(this)

    fun onConnect() = scope.launch {
        logger trace "Connected"
        packetsAdapter.launchPacketDataChannelConsumer()

        state = State.WaitingForHandshake
        if (server.eventBus.emit(SessionInitializedEvent(this@UranosSession)).cancelled) channel.close()
        else server.sessions.add(this@UranosSession)
    }

    fun onDisconnect(expected: Boolean) {
        scope.launch {
            if (state == State.Disconnected) return@launch
            if (!expected && currentProtocol != HandshakingProtocol && currentProtocol != StatusProtocol)
                logger trace "The client disconnected unexpectedly"

            packetsAdapter.stopProcessingIncomingPackets()
            coroutineContext.cancel(DisconnectedCancellationException())
            state = State.Disconnected
            server.sessions.remove(this@UranosSession)
        }
    }

    fun failAndDisconnectBecauseOfClient(throwable: Throwable) =
        failAndDisconnectBecauseOfClient("${throwable::class.java.name}: ${throwable.message}", throwable)

    internal fun failAndDisconnectBecauseOfClient(message: String, throwable: Throwable? = null) {
        val start = "The client caused a connection error"

        if (server.config.ignoreClientCausedErrors) {
            logger debug "The client caused a connection error: $message"
            // Does not send the Disconnect packet
            channel.close()
            onDisconnect(true)
        } else {
            if (throwable == null) logger.error("$start: $message")
            else logger.error(start, throwable)

            scope.launch {
                // Sends the Disconnect packet
                disconnect(internalReason = message)
            }
        }
    }

    fun enableEncryptionCodec(key: SecretKey) {
        logger trace "Enabling encryption codec"
        if (channel.pipeline().get("encryption") != null) return
        channel.pipeline().addBefore("framing", "encryption", EncryptionCodec(this, key))
    }

    fun enableCompressionCodec() {
        logger trace "Enabling compression codec"
        if (channel.pipeline().get("compression") != null) return

        channel.pipeline()
            .addAfter("framing", "compression", CompressionCodec(server.config.packetCompressionThreshold))
    }

    fun sendKeepAlivePacket() = send(OutgoingKeepAlivePacket(System.currentTimeMillis()))

    suspend fun tick(index: Int) {
        if (earlyPlayer != null && index % KEEP_ALIVE_PACKET_INTERVAL == 0L) sendKeepAlivePacket()

        packetsAdapter.tick()
    }

    companion object {
        val KEEP_ALIVE_PACKET_INTERVAL = msToTicks(1000)
    }
}
