package space.uranos.player

import space.uranos.chat.TextComponent
import space.uranos.entity.PlayerEntity
import space.uranos.entity.safeWorld
import space.uranos.net.UranosSession
import space.uranos.net.packet.play.ChunkDataPacket
import space.uranos.net.packet.play.ChunkLightDataPacket
import space.uranos.net.packet.play.PlayerInfoPacket
import space.uranos.net.packet.play.SelectedHotbarSlotPacket
import space.uranos.util.TickSynchronizationContainer
import space.uranos.util.numbers.validateParameterIsInRange
import space.uranos.world.Chunk
import space.uranos.world.VoxelLocation
import space.uranos.world.World
import java.util.UUID
import kotlin.math.abs

class UranosPlayer(
    override val session: UranosSession,
    override val name: String,
    override val uuid: UUID,
    override var gameMode: GameMode,
    override var settings: Player.Settings,
    override var reducedDebugInfo: Boolean,
    override var fieldOfView: Float,
    override var canFly: Boolean,
    override var flying: Boolean,
    override var flyingSpeed: Float,
    override var vulnerable: Boolean,
    override var compassTarget: VoxelLocation,
    selectedHotbarSlot: Int
) : Player {
    val container = TickSynchronizationContainer()

    override var selectedHotbarSlot by container.ifChanged(
        selectedHotbarSlot,
        { validateParameterIsInRange("selectedHotbarSlot", 0..8, it) }) {
        session.sendNow(SelectedHotbarSlotPacket(it))
    }

    override var playerListName by container.ifChanged<TextComponent?>(TextComponent of name) { value ->
        session.server.players.forEach {
            it.session.send(PlayerInfoPacket(PlayerInfoPacket.Action.UpdateDisplayName(mapOf(uuid to value))))
        }
    }

    init {
        this.selectedHotbarSlot = selectedHotbarSlot
        this.playerListName = null
    }

    override var currentlyViewedChunks = emptyList<Chunk>()

    override val entity: PlayerEntity = session.server.createPlayerEntity(this)

    suspend fun spawnInitially(world: World) {
        session.server.entities.forEach { if (it.visibleToNewPlayers && it != entity && it.world === world) it.viewers.add(this) }
        entity.setWorld(world)
        updateCurrentlyViewedChunks()
        sendChunksAndLight()
    }

    /**
     * Sets [currentlyViewedChunks] to all chunks in the view distance.
     */
    private fun updateCurrentlyViewedChunks() {
        val (centerX, centerZ) = Chunk.Key.from(entity.position.toVoxelLocation())

        val edgeLength = settings.viewDistance + 1

        currentlyViewedChunks = buildList(edgeLength * edgeLength) {
            for (x in (centerX - settings.viewDistance)..(centerX + settings.viewDistance)) {
                for (z in (centerZ - settings.viewDistance)..(centerZ + settings.viewDistance)) {
                    add(entity.safeWorld.getChunk(Chunk.Key(x, z)))
                }
            }
        }
    }

    private suspend fun sendChunksAndLight() {
        val chunks = currentlyViewedChunks.sortedBy { abs(it.key.x) + abs(it.key.z) }
        chunks.forEach { session.sendNow(ChunkLightDataPacket(it.key, it.getLightData(this))) }
        chunks.forEach { session.sendNow(ChunkDataPacket(it.key, it.getData(this))) }
    }
}
