package space.uranos.plugin

import space.uranos.UranosServer
import space.uranos.logging.Logger
import space.uranos.util.pluralize
import space.uranos.util.pluralizeWithCount
import java.net.URL
import java.net.URLClassLoader

class UranosPluginManager(private val server: UranosServer) : PluginManager {
    private val logger = Logger("PluginManager")

    override val plugins = mutableListOf<Plugin>()

    fun loadAll() {
        val pluginsDir = server.serverDirectory.resolve("plugins")
        pluginsDir.mkdirs()
        val files = pluginsDir.listFiles { it ->
            it.isFile && it.extension == "jar"
        }!!

        if (files.isEmpty()) {
            logger warn "No plugins found"
            return
        } else logger info "Loading ${pluralizeWithCount("plugin", files.size)}..."

        plugins.addAll(files.mapNotNull { file ->
            var error: LoadError?

            try {
                val loader = URLClassLoader(arrayOf(file.toURI().toURL()))
                val resource: URL? = loader.findResource(PLUGIN_CLASS_FILENAME)

                if (resource == null) error = LoadError.NoPluginClassFile
                else {
                    val pluginClassName = resource.readText().trim()

                    @Suppress("UNCHECKED_CAST")
                    val pluginClass = loader.loadClass(pluginClassName) as Class<Plugin>
                    return@mapNotNull pluginClass.getDeclaredConstructor().newInstance()
                }
            } catch (e: ClassNotFoundException) {
                error = LoadError.PluginClassNotFound(e.message!!)
            } catch (e: ClassCastException) {
                error = LoadError.NotExtending
            } catch (e: InstantiationException) {
                error = when (e.cause) {
                    is NoSuchMethodException -> LoadError.InvalidConstructor
                    else -> LoadError.Unknown(e)
                }
            } catch (e: IllegalAccessException) {
                error = LoadError.InvalidConstructor
            } catch (e: Exception) {
                error = LoadError.Unknown(e)
            }

            when {
                error is LoadError.Unknown -> logger.error("Loading ${file.name} failed", error.exception)
                error != null -> logger error "Loading ${file.name} failed: " + when (error) {
                    is LoadError.NotExtending -> "The plugin class must extend space.uranos.plugin.Plugin"
                    is LoadError.PluginClassNotFound -> "The plugin class '${error.name}' could not be found"
                    is LoadError.InvalidConstructor -> "The plugin class must have a public parameterless constructor"
                    is LoadError.NoPluginClassFile ->
                        "The plugin does not contain a plugin_class.txt file specifying the plugin class"

                    else -> "" // never happens
                }
            }

            null
        })

        logger info "Successfully loaded ${plugins.size}/${files.size} " +
                "${pluralize("plugin", plugins.size)}: " +
                plugins.joinToString { "${it.meta.name} (${it.meta.version})" }
    }

    sealed class LoadError {
        data class Unknown(val exception: Exception) : LoadError()
        data class PluginClassNotFound(val name: String) : LoadError()
        object InvalidConstructor : LoadError()
        object NotExtending : LoadError()
        object NoPluginClassFile : LoadError()
    }

    suspend fun enableAll() {
        plugins.forEach { it.onEnable() }
    }

    companion object {
        const val PLUGIN_CLASS_FILENAME = "plugin_class.txt"
    }
}
