package space.uranos.util

import com.squareup.moshi.Json
import com.squareup.moshi.JsonClass
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.withContext
import okhttp3.Cache
import okhttp3.CacheControl
import okhttp3.OkHttpClient
import okhttp3.Request
import space.uranos.Uranos
import java.util.UUID

object AuthenticationHelper {
    private const val URL = "https://sessionserver.mojang.com/session/minecraft/hasJoined"

    private val httpClient = OkHttpClient.Builder()
        .cache(
            Cache(
                directory = Uranos.cacheDirectory.resolve("auth"),
                maxSize = 50L * 1024L * 1024L // 10 MiB
            )
        )
        .build()

    suspend fun authenticate(hash: String, username: String): AuthenticationResult {
        val request = Request.Builder()
            .url(URL + "?serverId=${hash}&username=$username")
            .get()
            .cacheControl(CacheControl.Builder().immutable().build())
            .build()

        val response = httpClient.newCall(request).await()

        if (!response.isSuccessful)
            throw AuthenticationException("The response to the session server was not successful.")

        return withContext(Dispatchers.IO) {
            @Suppress("BlockingMethodInNonBlockingContext")
            moshi.adapter(AuthenticationResult::class.java).fromJson(response.body!!.source())!!
        }
    }
}

@JsonClass(generateAdapter = true)
data class AuthenticationResult(
    @Json(name = "id") @WithoutHyphens val uuid: UUID,
    @Json(name = "name") val username: String,
    val properties: List<Property>
) {
    @JsonClass(generateAdapter = true)
    data class Property(
        val name: String,
        val value: String,
        val signature: String
    )
}

class AuthenticationException(message: String) : Exception(message)
