package space.uranos.event

import kotlinx.coroutines.runBlocking
import org.junit.jupiter.api.BeforeEach
import org.junit.jupiter.api.Test
import space.uranos.Uranos
import space.uranos.UranosServer
import strikt.api.expectThat
import strikt.api.expectThrows
import strikt.assertions.isEqualTo

class EventBusTest {
    class TestEvent: Event() {
        private var count = 0

        fun test(expectedCount: Int) {
            expectThat(count).isEqualTo(expectedCount)
            count++
        }
    }

    object Position1 : EventHandlerPosition()
    object Position2 : EventHandlerPosition()
    object Position3 : EventHandlerPosition()
    object Position4 : EventHandlerPosition()
    object Position5 : EventHandlerPosition()
    object Position6 : EventHandlerPosition()

    @BeforeEach
    fun initServer() {
        UranosServer.setServerInstance(UranosServer())
    }

    @Test
    fun `handler positions cannot be inserted more than once`() {
        Uranos.eventHandlerPositions.insert(Position1)

        expectThrows<IllegalArgumentException> { Uranos.eventHandlerPositions.insert(Position1) }
        expectThrows<IllegalArgumentException> { Uranos.eventHandlerPositions.insertAfter(EventHandlerPosition.FIRST, Position1) }
        expectThrows<IllegalArgumentException> { Uranos.eventHandlerPositions.insertBefore(EventHandlerPosition.LAST, Position1) }
    }

    @Test
    fun `handler positions are inserted where they should be`() {
        val expectedOrder = listOf(
            EventHandlerPosition.FIRST,
            Position2,
            Position4,
            Position1,
            Position5,
            EventHandlerPosition.NORMAL,
            Position3,
            Position6,
            EventHandlerPosition.LAST
        )

        val actualOrder = (Uranos.eventHandlerPositions as UranosEventHandlerPositionManager).positions
        Uranos.eventHandlerPositions.insertAfter(EventHandlerPosition.FIRST, Position1, Position5)
        Uranos.eventHandlerPositions.insertBefore(Position1, Position2, Position4)
        Uranos.eventHandlerPositions.insert(Position3, Position6)

        actualOrder.forEachIndexed { index, position ->
            expectThat(index).isEqualTo(expectedOrder.indexOf(position))
        }
    }

    @Test
    fun `handler positions cannot be inserted before the first or after the last`() {
        expectThrows<IllegalArgumentException> {
            Uranos.eventHandlerPositions.insertBefore(EventHandlerPosition.FIRST, Position1)
        }

        expectThrows<IllegalArgumentException> {
            Uranos.eventHandlerPositions.insertAfter(EventHandlerPosition.LAST, Position1)
        }
    }

    @Test
    fun `handlers are invoked in the right order`() {
        val order = listOf(
            EventHandlerPosition.FIRST,
            Position1,
            Position2,
            EventHandlerPosition.NORMAL,
            Position3,
            EventHandlerPosition.LAST
        )

        Uranos.eventHandlerPositions.insertAfter(EventHandlerPosition.FIRST, Position1)
        Uranos.eventHandlerPositions.insertBefore(EventHandlerPosition.NORMAL, Position2)
        Uranos.eventHandlerPositions.insert(Position3)

        order.shuffled().forEach { pos -> Uranos.eventBus.on<TestEvent>(pos) { it.test(order.indexOf(pos)) } }

        runBlocking { Uranos.eventBus.emit(TestEvent()) }
    }
}
